# -*- coding: utf-8 -*-
"This module defines rules and algorithms for generating C++ code."

# Copyright (C) 2009-2016 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kristian B. Oelgaard 2011
# Modified by Marie E. Rognes 2010
# Modified by Martin Sandve Alnæs 2013-2016

# Python modules
import re
import numpy
import platform
from six import string_types
from six.moves import zip

# UFL modules
from ufl import custom_integral_types

# FFC modules
from ffc.log import debug, error


# Default precision for formatting floats
default_precision = numpy.finfo("double").precision + 1  # == 16


# ufc class names

def make_classname(prefix, basename, signature):
    pre = prefix.lower() + "_" if prefix else ""
    sig = str(signature).lower()
    return "%s%s_%s" % (pre, basename, sig)


def make_integral_classname(prefix, integral_type, form_id, subdomain_id):
    basename = "%s_integral_%s" % (integral_type, str(form_id).lower())
    return make_classname(prefix, basename, subdomain_id)


# Mapping of restrictions
_fixed_map = {None: "", "+": "_0", "-": "_1"}
_choose_map = lambda r: _fixed_map[r] if r in _fixed_map else "_%s" % str(r)

# FIXME: MSA: Using a dict to collect functions in a namespace is weird
#             and makes the code harder to follow, change to a class
#             with member functions instead!
# FIXME: KBO: format is a builtin_function, i.e., we should use a different name.
# Formatting rules
format = {}

# Program flow
format.update({
    "return": lambda v: "return %s;" % str(v),
    "grouping": lambda v: "(%s)" % v,
    "block": lambda v: "{%s}" % v,
    "block begin": "{",
    "block end": "}",
    "list": lambda v: format["block"](format["list separator"].join([str(l) for l in v])),
    "switch": lambda v, cases, default=None, numbers=None: _generate_switch(v, cases, default, numbers),
    "exception": lambda v: "throw std::runtime_error(\"%s\");" % v,
    "warning": lambda v: 'std::cerr << "*** FFC warning: " << "%s" << std::endl;' % v,
    "comment": lambda v: "// %s" % v,
    "if": lambda c, v: "if (%s)\n{\n%s\n}\n" % (c, v),
    "loop": lambda i, j, k: "for (unsigned int %s = %s; %s < %s; %s++)" % (i, j, i, k, i),
    "generate loop": lambda v, w, _indent=0: _generate_loop(v, w, _indent),
    "is equal": " == ",
    "not equal": " != ",
    "less than": " < ",
    "greater than": " > ",
    "less equal": " <= ",
    "greater equal": " >= ",
    "and": " && ",
    "or": " || ",
    "not": lambda v: "!(%s)" % v,
    "do nothing": "// Do nothing"
})

# Declarations
format.update({
    "declaration": lambda t, n, v=None: _declaration(t, n, v),
    "float declaration": "double",
    "int declaration": "int",
    "uint declaration": "unsigned int",
    "static const uint declaration": "static const unsigned int",
    "static const float declaration": "static const double",
    "vector table declaration": "std::vector< std::vector<double> >",
    "double array declaration": "double*",
    "const double array declaration": "const double*",
    "const float declaration": lambda v, w: "const double %s = %s;" % (v, w),
    "const uint declaration": lambda v, w: "const unsigned int %s = %s;" % (v, w),
    "dynamic array": lambda t, n, s: "%s *%s = new %s[%s];" % (t, n, t, s),
    "static array": lambda t, n, s: "static %s %s[%d];" % (t, n, s),
    "fixed array": lambda t, n, s: "%s %s[%d];" % (t, n, s),
    "delete dynamic array": lambda n, s=None: _delete_array(n, s),
    "create foo": lambda v: "new %s()" % v,
    "create factory": lambda v: "create_%s()" % v
})

# Mathematical operators
format.update({
    "add": lambda v: " + ".join(v),
    "iadd": lambda v, w: "%s += %s;" % (str(v), str(w)),
    "sub": lambda v: " - ".join(v),
    "neg": lambda v: "-%s" % v,
    "mul": lambda v: "*".join(v),
    "imul": lambda v, w: "%s *= %s;" % (str(v), str(w)),
    "div": lambda v, w: "%s/%s" % (str(v), str(w)),
    "inverse": lambda v: "(1.0/%s)" % v,
    "std power": lambda base, exp: "std::pow(%s, %s)" % (base, exp),
    "exp": lambda v: "std::exp(%s)" % str(v),
    "ln": lambda v: "std::log(%s)" % str(v),
    "cos": lambda v: "std::cos(%s)" % str(v),
    "sin": lambda v: "std::sin(%s)" % str(v),
    "tan": lambda v: "std::tan(%s)" % str(v),
    "cosh": lambda v: "std::cosh(%s)" % str(v),
    "sinh": lambda v: "std::sinh(%s)" % str(v),
    "tanh": lambda v: "std::tanh(%s)" % str(v),
    "acos": lambda v: "std::acos(%s)" % str(v),
    "asin": lambda v: "std::asin(%s)" % str(v),
    "atan": lambda v: "std::atan(%s)" % str(v),
    "atan_2": lambda v1, v2: "std::atan2(%s,%s)" % (str(v1), str(v2)),
    "erf": lambda v: "erf(%s)" % str(v),
    "bessel_i": lambda v, n: "boost::math::cyl_bessel_i(%s, %s)" % (str(n), str(v)),
    "bessel_j": lambda v, n: "boost::math::cyl_bessel_j(%s, %s)" % (str(n), str(v)),
    "bessel_k": lambda v, n: "boost::math::cyl_bessel_k(%s, %s)" % (str(n), str(v)),
    "bessel_y": lambda v, n: "boost::math::cyl_neumann(%s, %s)" % (str(n), str(v)),
    "absolute value": lambda v: "std::abs(%s)" % str(v),
    "min value": lambda l, r: "std::min(%s, %s)" % (str(l), str(r)),
    "max value": lambda l, r: "std::max(%s, %s)" % (str(l), str(r)),
    "sqrt": lambda v: "std::sqrt(%s)" % str(v),
    "addition": lambda v: _add(v),
    "multiply": lambda v: _multiply(v),
    "power": lambda base, exp: _power(base, exp),
    "inner product": lambda v, w: _inner_product(v, w),
    "assign": lambda v, w: "%s = %s;" % (v, str(w)),
    "component": lambda v, k: _component(v, k)
})

# Formatting used in tabulate_tensor
format.update({
    "geometry tensor": lambda j, a: "G%d_%s" % (j, "_".join(["%d" % i for i in a]))
})

# Geometry related variable names (from code snippets).
format.update({
    "entity index": "entity_indices",
    "num entities": "num_global_entities",
    "cell": lambda s: "ufc::shape::%s" % s,
    "J": lambda i, j, m, n: "J[%d]" % _flatten(i, j, m, n),
    "inv(J)": lambda i, j, m, n: "K[%d]" % _flatten(i, j, m, n),
    "det(J)": lambda r=None: "detJ%s" % _choose_map(r),
    "cell volume": lambda r=None: "volume%s" % _choose_map(r),
    "circumradius": lambda r=None: "circumradius%s" % _choose_map(r),
    "facet area": "facet_area",
    "min facet edge length": lambda r: "min_facet_edge_length",
    "max facet edge length": lambda r: "max_facet_edge_length",
    "scale factor": "det",
    "transform": lambda t, i, j, m, n, r: _transform(t, i, j, m, n, r),
    "normal component": lambda r, j: "n%s%s" % (_choose_map(r), j),
    "x coordinate": "X",
    "y coordinate": "Y",
    "z coordinate": "Z",
    "ip coordinates": lambda i, j: "X%d[%d]" % (i, j),
    "affine map table": lambda i, j: "FEA%d_f%d" % (i, j),
    "coordinate_dofs": lambda r=None: "coordinate_dofs%s" % _choose_map(r)
})

# UFC function arguments and class members (names)
format.update({
    "element tensor": lambda i: "A[%s]" % i,
    "element tensor term": lambda i, j: "A%d[%s]" % (j, i),
    "coefficient": lambda j, k: format["component"]("w", [j, k]),
    "argument basis num": "i",
    "argument derivative order": "n",
    "argument values": "values",
    "argument coordinates": "dof_coordinates",
    "facet": lambda r: "facet%s" % _choose_map(r),
    "vertex": "vertex",
    "argument axis": "i",
    "argument dimension": "d",
    "argument entity": "i",
    "member global dimension": "_global_dimension",
    "argument dofs": "dofs",
    "argument dof num": "i",
    "argument dof values": "dof_values",
    "argument vertex values": "vertex_values",
    "argument sub": "i",  # sub element
    "argument subdomain": "subdomain_id",  # sub domain
})

# Formatting used in evaluatedof.
format.update({
    "dof vals": "vals",
    "dof result": "result",
    "dof X": lambda i: "X_%d" % i,
    "dof D": lambda i: "D_%d" % i,
    "dof W": lambda i: "W_%d" % i,
    "dof copy": lambda i: "copy_%d" % i,
    "dof physical coordinates": "y"
})


# Formatting used in evaluate_basis, evaluate_basis_derivatives and quadrature
# code generators.
format.update({
    # evaluate_basis and evaluate_basis_derivatives
    "tmp value": lambda i: "tmp%d" % i,
    "tmp ref value": lambda i: "tmp_ref%d" % i,
    "local dof": "dof",
    "basisvalues": "basisvalues",
    "coefficients": lambda i: "coefficients%d" % (i),
    "num derivatives": lambda t_or_g: "num_derivatives" + t_or_g,
    "derivative combinations": lambda t_or_g: "combinations" + t_or_g,
    "transform matrix": "transform",
    "transform Jinv": "Jinv",
    "dmats": lambda i: "dmats%s" % (i),
    "dmats old": "dmats_old",
    "reference derivatives": "derivatives",
    "dof values": "dof_values",
    "dof map if":                 lambda i, j: "%d <= %s && %s <= %d"\
                                  % (i, format["argument basis num"], format["argument basis num"], j),
    "dereference pointer": lambda n: "*%s" % n,
    "reference variable": lambda n: "&%s" % n,
    "call basis": lambda i, s: "_evaluate_basis(%s, %s, x, coordinate_dofs, cell_orientation);" % (i, s),
    "call basis_all": "_evaluate_basis_all(values, x, coordinate_dofs, cell_orientation);",
    "call basis_derivatives": lambda i, s: "_evaluate_basis_derivatives(%s, n, %s, x, coordinate_dofs, cell_orientation);" % (i, s),
    "call basis_derivatives_all": lambda i, s: "_evaluate_basis_derivatives_all(n, %s, x, coordinate_dofs, cell_orientation);" % s,

    # quadrature code generators
    "integration points": "ip",
    "first free index": "j",
    "second free index": "k",
    "geometry constant": lambda i: "G[%d]" % i,
    "ip constant": lambda i: "I[%d]" % i,
    "basis constant": lambda i: "B[%d]" % i,
    "conditional": lambda i: "C[%d]" % i,
    "evaluate conditional": lambda i, j, k: "(%s) ? %s : %s" % (i, j, k),
              #    "geometry constant":   lambda i: "G%d" % i,
              #    "ip constant":         lambda i: "I%d" % i,
              #    "basis constant":      lambda i: "B%d" % i,
              "function value": lambda i: "F%d" % i,
              "nonzero columns": lambda i: "nzc%d" % i,
              "weight": lambda i: "W" if i is None else "W%d" % (i),
              "psi name": lambda c, et, e, co, d, a: _generate_psi_name(c, et, e, co, d, a),
              # both
              "free indices": ["r", "s", "t", "u"],
              "matrix index": lambda i, j, range_j: _matrix_index(i, str(j), str(range_j)),
              "quadrature point": lambda i, gdim: "quadrature_points + %s*%d" % (i, gdim),
              "facet_normal_custom": lambda gdim: _generate_facet_normal_custom(gdim),
              })

# Misc
format.update({
    "bool": lambda v: {True: "true", False: "false"}[v],
    "str": lambda v: "%s" % v,
    "int": lambda v: "%d" % v,
    "list separator": ", ",
    "block separator": ",\n",
    "new line": "\\\n",
    "tabulate tensor": lambda m: _tabulate_tensor(m),
})

# Code snippets
from ffc.codesnippets import *

format.update({
    "compute_jacobian": lambda tdim, gdim, r=None:
                                compute_jacobian[tdim][gdim] % {"restriction": _choose_map(r)},
    "compute_jacobian_inverse": lambda tdim, gdim, r=None:
                                compute_jacobian_inverse[tdim][gdim] % {"restriction": _choose_map(r)},
    "orientation": lambda tdim, gdim, r=None: orientation_snippet % {"restriction": _choose_map(r)} if tdim != gdim else "",
    "facet determinant": lambda tdim, gdim, r=None: facet_determinant[tdim][gdim] % {"restriction": _choose_map(r)},
    "fiat coordinate map": lambda cell, gdim: fiat_coordinate_map[cell][gdim],
    "generate normal": lambda tdim, gdim, i: _generate_normal(tdim, gdim, i),
    "generate cell volume": lambda tdim, gdim, i, r=None: _generate_cell_volume(tdim, gdim, i, r),
    "generate circumradius": lambda tdim, gdim, i, r=None: _generate_circumradius(tdim, gdim, i, r),
    "generate facet area": lambda tdim, gdim: facet_area[tdim][gdim],
    "generate min facet edge length": lambda tdim, gdim, r=None: min_facet_edge_length[tdim][gdim] % {"restriction": _choose_map(r)},
    "generate max facet edge length": lambda tdim, gdim, r=None: max_facet_edge_length[tdim][gdim] % {"restriction": _choose_map(r)},
    "generate ip coordinates": lambda g, t, num_ip, name, ip, r=None: (ip_coordinates[t][g][0], ip_coordinates[t][g][1] %
                                                                       {"restriction": _choose_map(r), "ip": ip, "name": name, "num_ip": num_ip}),
    "scale factor snippet": scale_factor,
    "map onto physical": map_onto_physical,
    "evaluate basis snippet": eval_basis,
    "combinations": combinations_snippet,
    "transform snippet": transform_snippet,
    "evaluate function": evaluate_f,
    "ufc comment": comment_ufc,
    "dolfin comment": comment_dolfin,
    "header_h": header_h,
    "header_c": header_c,
    "footer": footer,
    "eval_basis_decl": eval_basis_decl,
    "eval_basis_init": eval_basis_init,
    "eval_basis": eval_basis,
    "eval_basis_copy": eval_basis_copy,
    "eval_derivs_decl": eval_derivs_decl,
    "eval_derivs_init": eval_derivs_init,
    "eval_derivs": eval_derivs,
    "eval_derivs_copy": eval_derivs_copy,
    "extract_cell_coordinates": lambda offset, r: "const double* coordinate_dofs_%d = coordinate_dofs + %d;" % (r, offset)
})

# Helper functions for formatting


def _declaration(type, name, value=None):
    if value is None:
        return "%s %s;" % (type, name)
    return "%s %s = %s;" % (type, name, str(value))


def _component(var, k):
    if not isinstance(k, (list, tuple)):
        k = [k]
    return "%s" % var + "".join("[%s]" % str(i) for i in k)


def _delete_array(name, size=None):
    if size is None:
        return "delete [] %s;" % name
    f_r = format["free indices"][0]
    code = format["generate loop"](["delete [] %s;" % format["component"](name, f_r)], [(f_r, 0, size)])
    code.append("delete [] %s;" % name)
    return "\n".join(code)


def _multiply(factors):
    """
    Generate string multiplying a list of numbers or strings.  If a
    factor is zero, the whole product is zero. Any factors equal to
    one are ignored.
    """

    # FIXME: This could probably be way more robust and elegant.

    cpp_str = format["str"]
    non_zero_factors = []
    for f in factors:

        # Round-off if f is smaller than epsilon
        if isinstance(f, (int, float)):
            if abs(f) < format["epsilon"]:
                return cpp_str(0)
            if abs(f - 1.0) < format["epsilon"]:
                continue

        # Convert to string
        f = cpp_str(f)

        # Return zero if any factor is zero
        if f == "0":
            return cpp_str(0)

        # If f is 1, don't add it to list of factors
        if f == "1":
            continue

        # If sum-like, parentheseze factor
        if "+" in f or "-" in f:
            f = "(%s)" % f

        non_zero_factors += [f]

    if len(non_zero_factors) == 0:
        return cpp_str(1.0)

    return "*".join(non_zero_factors)


def _add(terms):
    "Generate string summing a list of strings."

    # FIXME: Subtract absolute value of negative numbers
    result = " + ".join([str(t) for t in terms if (str(t) != "0")])
    if result == "":
        return format["str"](0)
    return result


def _power(base, exponent):
    "Generate code for base^exponent."
    if exponent >= 0:
        return _multiply(exponent * (base,))
    else:
        return "1.0 / (%s)" % _power(base, -exponent)


def _inner_product(v, w):
    "Generate string for v[0]*w[0] + ... + v[n]*w[n]."

    # Check that v and w have same length
    assert(len(v) == len(w)), "Sizes differ in inner-product!"

    # Special case, zero terms
    if len(v) == 0:
        return format["float"](0)

    # Straightforward handling when we only have strings
    if isinstance(v[0], string_types):
        return _add([_multiply([v[i], w[i]]) for i in range(len(v))])

    # Fancy handling of negative numbers etc
    result = None
    eps = format["epsilon"]
    add = format["add"]
    sub = format["sub"]
    neg = format["neg"]
    mul = format["mul"]
    fl = format["float"]
    for (c, x) in zip(v, w):
        if result:
            if abs(c - 1.0) < eps:
                result = add([result, x])
            elif abs(c + 1.0) < eps:
                result = sub([result, x])
            elif c > eps:
                result = add([result, mul([fl(c), x])])
            elif c < -eps:
                result = sub([result, mul([fl(-c), x])])
        else:
            if abs(c - 1.0) < eps:
                result = x
            elif abs(c + 1.0) < eps:
                result = neg(x)
            elif c > eps:
                result = mul([fl(c), x])
            elif c < -eps:
                result = neg(mul([fl(-c), x]))

    return result


def _transform(type, i, j, m, n, r):
    map_name = {"J": "J", "JINV": "K"}[type] + _choose_map(r)
    return (map_name + "[%d]") % _flatten(i, j, m, n)

# FIXME: Input to _generate_switch should be a list of tuples (i, case)


def _generate_switch(variable, cases, default=None, numbers=None):
    "Generate switch statement from given variable and cases"

    # Special case: no cases and no default
    if len(cases) == 0 and default is None:
        return format["do nothing"]
    elif len(cases) == 0:
        return default

    # Special case: one case and no default
    if len(cases) == 1 and default is None:
        return cases[0]

    # Create numbers for switch
    if numbers is None:
        numbers = list(range(len(cases)))

    # Create switch
    code = "switch (%s)\n{\n" % variable
    for (i, case) in enumerate(cases):
        code += "case %d:\n  {\n  %s\n    break;\n  }\n" % (numbers[i], indent(case, 2))
    code += "}\n"

    # Default value
    if default:
        code += "\n" + default

    return code


def _tabulate_tensor(vals):
    "Tabulate a multidimensional tensor. (Replace tabulate_matrix and tabulate_vector)."

    # Prefetch formats to speed up code generation
    f_block = format["block"]
    f_list_sep = format["list separator"]
    f_block_sep = format["block separator"]
    # FIXME: KBO: Change this to "float" once issue in set_float_formatting is fixed.
    f_float = format["floating point"]
    f_epsilon = format["epsilon"]

    # Create numpy array and get shape.
    tensor = numpy.array(vals)
    shape = numpy.shape(tensor)
    if len(shape) == 1:
        # Create zeros if value is smaller than tolerance.
        values = []
        for v in tensor:
            if abs(v) < f_epsilon:
                values.append(f_float(0.0))
            else:
                values.append(f_float(v))
        # Format values.
        return f_block(f_list_sep.join(values))
    elif len(shape) > 1:
        return f_block(f_block_sep.join([_tabulate_tensor(tensor[i]) for i in range(shape[0])]))
    else:
        error("Not an N-dimensional array:\n%s" % tensor)


def _generate_loop(lines, loop_vars, _indent):
    "This function generates a loop over a vector or matrix."

    # Prefetch formats to speed up code generation.
    f_loop = format["loop"]
    f_begin = format["block begin"]
    f_end = format["block end"]
    f_comment = format["comment"]

    if not loop_vars:
        return lines

    code = []
    for ls in loop_vars:
        # Get index and lower and upper bounds.
        index, lower, upper = ls
        # Loop index.
        code.append(indent(f_loop(index, lower, upper), _indent))
        code.append(indent(f_begin, _indent))

        # Increase indentation.
        _indent += 2

        # If this is the last loop, write values.
        if index == loop_vars[-1][0]:
            for l in lines:
                code.append(indent(l, _indent))

    # Decrease indentation and write end blocks.
    indices = [var[0] for var in loop_vars]
    indices.reverse()
    for index in indices:
        _indent -= 2
        code.append(indent(f_end + " " + f_comment("end loop over '%s'" % index), _indent))

    return code


def _matrix_index(i, j, range_j):
    "Map the indices in a matrix to an index in an array i.e., m[i][j] -> a[i*range(j)+j]"
    if i == 0:
        access = j
    elif i == 1:
        access = format["add"]([range_j, j])
    else:
        irj = format["mul"]([format["str"](i), range_j])
        access = format["add"]([irj, j])
    return access


def _generate_psi_name(counter, entity_type, entity, component, derivatives, avg):
    """Generate a name for the psi table of the form:
    FE#_f#_v#_C#_D###_A#, where '#' will be an integer value.

    FE  - is a simple counter to distinguish the various bases, it will be
          assigned in an arbitrary fashion.

    f   - denotes facets if applicable, range(element.num_facets()).

    v   - denotes vertices if applicable, range(num_vertices).

    C   - is the component number if any (flattened in the case of tensor valued functions)

    D   - is the number of derivatives in each spatial direction if any.
          If the element is defined in 3D, then D012 means d^3(*)/dydz^2.

    A   - denotes averaged over cell (AC) or facet (AF)
    """

    name = "FE%d" % counter

    if entity_type == "facet":
        name += "_f%d" % entity
    elif entity_type == "vertex":
        name += "_v%d" % entity

    if component != () and component != []:
        name += "_C%d" % component

    if any(derivatives):
        name += "_D" + "".join(map(str, derivatives))

    if avg == "cell":
        name += "_AC"
    elif avg == "facet":
        name += "_AF"

    return name


def _generate_normal(tdim, gdim, integral_type, reference_normal=False):
    "Generate code for computing normal"

    # Choose snippets
    direction = normal_direction[tdim][gdim]

    assert (gdim in facet_normal[tdim]),\
        "Facet normal not yet implemented for this tdim/gdim combo"
    normal = facet_normal[tdim][gdim]

    # Choose restrictions
    if integral_type == "exterior_facet":
        code = direction % {"restriction": "", "facet": "facet"}
        code += normal % {"direction": "", "restriction": ""}
    elif integral_type == "interior_facet":
        code = direction % {"restriction": _choose_map("+"), "facet": "facet_0"}
        code += normal % {"direction": "", "restriction": _choose_map("+")}
        code += normal % {"direction": "!", "restriction": _choose_map("-")}
    else:
        error("Unsupported integral_type: %s" % str(integral_type))
    return code


def _generate_facet_normal_custom(gdim):
    "Generate code for setting facet normal in custom integrals"
    code = format["comment"]("Set facet normal components for current quadrature point\n")
    for i in range(gdim):
        code += "const double n_0%d =   facet_normals[%d*ip + %d];\n" % (i, gdim, i)
        code += "const double n_1%d = - facet_normals[%d*ip + %d];\n" % (i, gdim, i)
    return code


def _generate_cell_volume(tdim, gdim, integral_type, r=None):
    "Generate code for computing cell volume."

    # Choose snippets
    volume = cell_volume[tdim][gdim]

    # Choose restrictions
    if integral_type in ("cell", "exterior_facet"):
        code = volume % {"restriction": ""}
    elif integral_type == "interior_facet":
        code = volume % {"restriction": _choose_map("+")}
        code += volume % {"restriction": _choose_map("-")}
    elif integral_type in custom_integral_types:
        code = volume % {"restriction": _choose_map(r)}
    else:
        error("Unsupported integral_type: %s" % str(integral_type))
    return code


def _generate_circumradius(tdim, gdim, integral_type, r=None):
    "Generate code for computing a cell's circumradius."

    # Choose snippets
    radius = circumradius[tdim][gdim]

    # Choose restrictions
    if integral_type in ("cell", "exterior_facet", "vertex"):
        code = radius % {"restriction": ""}
    elif integral_type == "interior_facet":
        code = radius % {"restriction": _choose_map("+")}
        code += radius % {"restriction": _choose_map("-")}
    elif integral_type in custom_integral_types:
        code = radius % {"restriction": _choose_map(r)}
    else:
        error("Unsupported integral_type: %s" % str(integral_type))
    return code


def _flatten(i, j, m, n):
    return i * n + j

# Other functions


def indent(block, num_spaces):
    "Indent each row of the given string block with n spaces."
    indentation = " " * num_spaces
    return indentation + ("\n" + indentation).join(block.split("\n"))


def count_ops(code):
    "Count the number of operations in code (multiply-add pairs)."
    num_add = code.count(" + ") + code.count(" - ")
    num_multiply = code.count("*") + code.count("/")
    return (num_add + num_multiply) // 2


def set_float_formatting(precision):
    "Set floating point formatting based on precision."

    # Set default if not set
    if precision is None:
        precision = default_precision

    # Options for float formatting
    # f1     = "%%.%df" % precision
    # f2     = "%%.%de" % precision
    f1 = "%%.%dg" % precision
    f2 = "%%.%dg" % precision
    f_int = "%%.%df" % 1

    eps = eval("1e-%s" % precision)

    # Regular float formatting
    def floating_point_regular(v):
        if abs(v - round(v, 1)) < eps:
            return f_int % v
        elif abs(v) < 100.0:
            return f1 % v
        else:
            return f2 % v

    # Special float formatting on Windows (remove extra leading zero)
    def floating_point_windows(v):
        return floating_point_regular(v).replace("e-0", "e-").replace("e+0", "e+")

    # Set float formatting
    if platform.system() == "Windows":
        format["float"] = floating_point_windows
    else:
        format["float"] = floating_point_regular

    # FIXME: KBO: Remove once we agree on the format of 'f1'
    format["floating point"] = format["float"]

    # Set machine precision
    format["epsilon"] = 10.0 * eval("1e-%s" % precision)

    # Hack to propagate precision to uflacs internals...
    #import ffc.uflacs.language.format_value
    #ffc.uflacs.language.format_value.set_float_precision(precision)


def set_exception_handling(convert_exceptions_to_warnings):
    "Set handling of exceptions."
    if convert_exceptions_to_warnings:
        format["exception"] = format["warning"]


# Declarations to examine
types = [["double"],
         ["const", "double"],
         ["const", "double", "*", "const", "*"],
         ["int"],
         ["const", "int"],
         ["unsigned", "int"],
         ["bool"],
         ["const", "bool"],
         ["static", "unsigned", "int"],
         ["const", "unsigned", "int"]]

# Special characters and delimiters
special_characters = ["+", "-", "*", "/", "=", ".", " ", ";", "(", ")", "\\", "{", "}", "[", "]", "!"]


def remove_unused(code, used_set=set()):
    """
    Remove unused variables from a given C++ code. This is useful when
    generating code that will be compiled with gcc and parameters -Wall
    -Werror, in which case gcc returns an error when seeing a variable
    declaration for a variable that is never used.

    Optionally, a set may be specified to indicate a set of variables
    names that are known to be used a priori.
    """

    # Dictionary of (declaration_line, used_lines) for variables
    variables = {}

    # List of variable names (so we can search them in order)
    variable_names = []

    lines = code.split("\n")
    for (line_number, line) in enumerate(lines):
        # Exclude commented lines.
        if line[:2] == "//" or line[:3] == "///":
            continue

        # Split words
        words = [word for word in line.split(" ") if word != ""]

        # Remember line where variable is declared
        for type in [type for type in types if " ".join(type) in " ".join(words)]:  # Fewer matches than line below.
        # for type in [type for type in types if len(words) > len(type)]:
            variable_type = words[0:len(type)]
            variable_name = words[len(type)]

            # Skip special characters
            if variable_name in special_characters:
                continue

            # Test if any of the special characters are present in the variable name
            # If this is the case, then remove these by assuming that the 'real' name
            # is the first entry in the return list. This is implemented to prevent
            # removal of e.g. 'double array[6]' if it is later used in a loop as 'array[i]'
            if variable_type == type:

                # Create correct variable name (e.g. y instead of
                # y[2]) for variables with separators
                seps_present = [sep for sep in special_characters if sep in variable_name]
                if seps_present:
                    variable_name = sorted([variable_name.split(sep)[0] for sep in seps_present])
                    variable_name = variable_name[0]

                variables[variable_name] = (line_number, [])
                if variable_name not in variable_names:
                    variable_names += [variable_name]

        # Mark line for used variables
        for variable_name in variables:
            (declaration_line, used_lines) = variables[variable_name]
            if _variable_in_line(variable_name, line) and line_number > declaration_line:
                variables[variable_name] = (declaration_line, used_lines + [line_number])

    # Reverse the order of the variable names to catch variables used
    # only by variables that are removed
    variable_names.reverse()

    # Remove declarations that are not used
    removed_lines = []
    for variable_name in variable_names:
        (declaration_line, used_lines) = variables[variable_name]
        for line in removed_lines:
            if line in used_lines:
                used_lines.remove(line)
        if not used_lines and variable_name not in used_set:
            debug("Removing unused variable: %s" % variable_name)
            lines[declaration_line] = None  # KBO: Need to completely remove line for evaluate_basis* to work
            # lines[declaration_line] = "// " + lines[declaration_line]
            removed_lines += [declaration_line]
    return "\n".join([line for line in lines if line is not None])


def _variable_in_line(variable_name, line):
    "Check if variable name is used in line"
    if variable_name not in line:
        return False
    for character in special_characters:
        line = line.replace(character, "\\" + character)
    delimiter = "[" + ",".join(["\\" + c for c in special_characters]) + "]"
    return re.search(delimiter + variable_name + delimiter, line) is not None



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.reduce_operations import expand_operations, reduce_operations
from ffc.quadrature.symbolics import *
from ffc.quadrature.sumobj import _group_fractions
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])
from ffc.log import error, push_level, pop_level, CRITICAL


def test_Float_Operators():
    "Test binary operators"

    f0 = FloatValue(0.0)
    f2 = FloatValue(2.0)
    f3 = FloatValue(3.0)
    fm1 = FloatValue(-1.0)
    fm3 = FloatValue(-3.0)

    x = Symbol("x", GEO)
    y = Symbol("y", GEO)
    z = Symbol("z", GEO)

    p0 = Product([f2, x])
    p1 = Product([x, y])
    p2 = Product([f2, z])
    p3 = Product([y, x, z])
    p4 = Product([fm1, f2, x])

    S0 = Sum([p0, fm3])
    S1 = Sum([x, y])
    S2 = Sum([S1, fm3])
    S3 = Sum([p4, fm3])
    S4 = Sum([fm3, Product([fm1, Sum([x, y])])])

    F0 = Fraction(f2, y)
    F1 = Fraction(FloatValue(-1.5), x)
    F2 = Fraction(fm3, S1)

    SF0 = Sum([f3, F1])
    SF1 = Sum([f3, Product([fm1, F1])])

    # Test FloatValue '+'
    assert str(f2 + fm3) == str(fm1)
    assert str(f2 + fm3 + fm3 + f2 + f2) == str(f0)
    assert str(f0 + p0) == str(p0)
    assert str(fm3 + p0) == str(S0)
    assert str(fm3 + S1) == str(S2)
    assert str(f3 + F1) == str(SF0)

    # Test FloatValue '-'
    assert str(f2 - fm3) == str(FloatValue(5))
    assert str(f0 - p0) == str(p4)
    assert str(fm3 - p0) == str(S3)
    assert str(fm3 - S1) == str(S4)
    assert str(f3 - F1) == str(SF1)

    # Test FloatValue '*', only need one because all other cases are
    # handled by 'other'
    assert str(f2 * f2) == '%s' % format["float"](4)

    # Test FloatValue '/'
    assert str(fm3 / f2) == str(FloatValue(-1.5))
    assert str(f2 / y) == str(F0)
    assert str(fm3 / p0) == str(F1)
    assert str(fm3 / S1) == str(F2)

    # Silence output
    push_level(CRITICAL)
    with pytest.raises(Exception):
        truediv(f2, F0)
    with pytest.raises(Exception):
        truediv(f2, f0)
    with pytest.raises(Exception):
        truediv(f2, Product([f0, y]))
    pop_level()



# -*- coding: utf-8 -*-
# Copyright (C) 2010-2016 Kristian B. Oelgaard and Garth N. Wells
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
# Modified by Lizao Li, 2016

import pytest
import os
import numpy
import subprocess
import time

from ufl import FiniteElement, MixedElement
from ufl.utils.py23 import as_native_str
import ffc
from ffc.log import error
from ffc.mixedelement import MixedElement as FFCMixedElement
from ffc.fiatinterface import create_element, reference_cell

# Local imports
from cppcode import evaluate_basis_code_fiat

# Tolerances
tol = 1e-14
crit_tol = 1e-8
# crit_tol = 1e-16

# Some random points
random_points = {1: [(0.114,), (0.349,), (0.986,)],
                 2: [(0.114, 0.854), (0.349, 0.247), (0.986, 0.045)],
                 3: [(0.114, 0.854, 0.126), (0.349, 0.247, 0.457),
                     (0.986, 0.045, 0.127)]}


def xcomb(items, n):
    "Create n-tuples with combinations of items."
    if n == 0:
        yield []
    else:
        for i in range(len(items)):
            for cc in xcomb(items[:i] + items[i + 1:], n - 1):
                yield [items[i]] + cc


# Elements, supported by FFC and FIAT, and their supported shape and orders
# TODO: This is not used anywhere, was it before? What is the purpose?
single_elements = [{"family": "Lagrange",
                    "shapes": ["interval", "triangle", "tetrahedron"],
                    "orders": [1, 2, 3, 4]},
                   {"family": "Discontinuous Lagrange",
                    "shapes": ["interval", "triangle", "tetrahedron"],
                    "orders": [0, 1, 2, 3, 4]},
                   {"family": "Crouzeix-Raviart",
                    "shapes": ["triangle", "tetrahedron"],
                    "orders": [1]},
                   {"family": "Raviart-Thomas",
                    "shapes": ["triangle", "tetrahedron"],
                    "orders": [1, 2, 3]},
                   {"family": "Discontinuous Raviart-Thomas",
                    "shapes": ["triangle", "tetrahedron"],
                    "orders": [1, 2, 3]},
                   {"family": "Brezzi-Douglas-Marini",
                    "shapes": ["triangle", "tetrahedron"],
                    "orders": [1, 2, 3]},
                   {"family": "Brezzi-Douglas-Fortin-Marini",
                    "shapes": ["triangle"],
                    "orders": [2]},
                   {"family": "Nedelec 1st kind H(curl)",
                    "shapes": ["triangle", "tetrahedron"],
                    "orders": [1, 2, 3]},
                   {"family": "Nedelec 2nd kind H(curl)",
                    "shapes": ["triangle", "tetrahedron"],
                    "orders": [1, 2, 3]},
                   {"family": "Regge",
                    "shapes": ["triangle", "tetrahedron"],
                    "orders": [0, 1, 2, 3]},
                   {"family": "HHJ",
                    "shapes": ["triangle"],
                    "orders": [0, 1, 2, 3]}]


# Create some mixed elements
dg0_tri = FiniteElement("DG", "triangle", 0)
dg1_tri = FiniteElement("DG", "triangle", 1)
cg1_tri = FiniteElement("CG", "triangle", 1)
cr1_tri = FiniteElement("CR", "triangle", 1)
rt1_tri = FiniteElement("RT", "triangle", 1)
drt2_tri = FiniteElement("DRT", "triangle", 2)
bdm1_tri = FiniteElement("BDM", "triangle", 1)
ned1_tri = FiniteElement("N1curl", "triangle", 1)
reg0_tri = FiniteElement("Regge", "triangle", 0)
hhj0_tri = FiniteElement("HHJ", "triangle", 0)

dg0_tet = FiniteElement("DG", "tetrahedron", 0)
dg1_tet = FiniteElement("DG", "tetrahedron", 1)
cg1_tet = FiniteElement("CG", "tetrahedron", 1)
cr1_tet = FiniteElement("CR", "tetrahedron", 1)
rt1_tet = FiniteElement("RT", "tetrahedron", 1)
drt2_tet = FiniteElement("DRT", "tetrahedron", 2)
bdm1_tet = FiniteElement("BDM", "tetrahedron", 1)
ned1_tet = FiniteElement("N1curl", "tetrahedron", 1)
reg0_tet = FiniteElement("Regge", "tetrahedron", 0)

# Create combinations in pairs
mix_tri = [MixedElement(e) for e in xcomb([dg0_tri, dg1_tri, cg1_tri, cr1_tri,
                                           rt1_tri, drt2_tri, bdm1_tri,
                                           ned1_tri, reg0_tri, hhj0_tri], 2)]
mix_tet = [MixedElement(e) for e in xcomb([dg0_tet, dg1_tet, cg1_tet, cr1_tet,
                                           rt1_tet, drt2_tet, bdm1_tet,
                                           ned1_tet, reg0_tet], 2)]

mixed_elements = [MixedElement([dg0_tri] * 4),
                  MixedElement([cg1_tri] * 3),
                  MixedElement([bdm1_tri] * 2),
                  MixedElement([dg1_tri, cg1_tri, cr1_tri, rt1_tri, bdm1_tri,
                                ned1_tri]),
                  MixedElement([MixedElement([rt1_tri, cr1_tri]), cg1_tri,
                                ned1_tri]),
                  MixedElement([ned1_tri, dg1_tri, MixedElement([rt1_tri,
                                                                 cr1_tri])]),
                  MixedElement([drt2_tri, cg1_tri]),
                  MixedElement([dg0_tet] * 4), MixedElement([cg1_tet] * 3),
                  MixedElement([bdm1_tet] * 2),
                  MixedElement([dg1_tet, cg1_tet, cr1_tet, rt1_tet, bdm1_tet,
                                ned1_tet]),
                  MixedElement([MixedElement([rt1_tet, cr1_tet]), cg1_tet,
                                ned1_tet]),
                  MixedElement([ned1_tet, dg1_tet, MixedElement([rt1_tet,
                                                                 cr1_tet])]),
                  MixedElement([drt2_tet, cg1_tet]),
                  MixedElement([cg1_tet, cg1_tet, cg1_tet, reg0_tet]),
                  MixedElement([reg0_tet] * 2),
                  MixedElement([hhj0_tri, ned1_tri])] + mix_tri + mix_tet

# Collect all elements here
all_elements = mixed_elements


def compile_gcc_code(ufl_element, code):
    # Write code to file
    with open("evaluate_basis.cpp", "w") as f:
        f.write(code)

    # Get location of UFC file
    from ffc import get_include_path, get_ufc_cxx_flags
    ufc_include_path = get_include_path()
    ufc_cxx_flags = get_ufc_cxx_flags()

    # Compile c++ code
    c = "c++ -I{} {} -Wall -Werror -o evaluate_basis_test_code evaluate_basis.cpp".format(ufc_include_path, " ".join(ufc_cxx_flags))
    with open("compile.sh", "w") as f:
        f.write(c + "\n")
    subprocess.check_call(c, shell=True)


def run_code(ufl_element, deriv_order):
    "Compute values of basis functions for given element."

    # Run compiled code and get values
    c = ".%sevaluate_basis_test_code %d" % (os.path.sep, deriv_order)
    try:
        output = as_native_str(subprocess.check_output(c, shell=True))
    except subprocess.CalledProcessError as e:
        print("Could not run compile code for element: {}".format(str(ufl_element)))
        raise

    values = [[float(value) for value in line.strip().split(" ") if value] for line in output.strip().split("\n")]
    return numpy.array(values)


def get_element_name(ufl_element):
    "Extract relevant element name from header file."
    with open("test.h") as f:
        lines = f.readlines()

    signature = repr(ufl_element)
    name = None
    for e, l in enumerate(lines):
        if "class" in l and "finite_element" in l:
            name = l
        if signature in l:
            break
    if name is None:
        raise RuntimeError("No finite element class found")
    return name.split()[1][:-1]


def generate_element(ufl_element):
    "Create UFL form file with a single element in it and compile it with FFC"
    with open("test.ufl", "w") as f:
        f.write("element = " + repr(ufl_element))
    r = ffc.main(["test.ufl"])
    assert r == 0, "FFC compilation failed for element: {}.".format(str(ufl_element))


def matrix(points):
    return "{%s};" % ", ".join(["{%s}" % ", ".join([str(c) for c in p]) for p in points])


def to_fiat_tuple(comb, gdim):
    """Convert a list of combinations of derivatives to a fiat tuple of
    derivatives.  FIAT expects a list with the number of derivatives
    in each spatial direction.  E.g., in 2D: u_{xyy} --> [0, 1, 1] in
    FFC --> (1, 2) in FIAT.

    """
    new_comb = [0] * gdim
    if comb == []:
        return tuple(new_comb)
    for i in range(gdim):
        new_comb[i] = comb.count(i)
    return tuple(new_comb)


def compute_derivative_combinations(deriv_order, geo_dim):
    "Compute combinations of derivatives in spatial directions (like code snippet)."

    if deriv_order == 0:
        return [(0,) * geo_dim]

    num_derivatives = geo_dim**deriv_order
    combinations = [[0] * deriv_order for n in range(num_derivatives)]
    for i in range(1, num_derivatives):
        for k in range(i):
            j = deriv_order - 1
            while j + 1 > 0:
                j -= 1
                if combinations[i][j] + 1 > geo_dim - 1:
                    combinations[i][j] = 0
                else:
                    combinations[i][j] += 1
                    break
    # Convert to fiat tuples.
    for i in range(num_derivatives):
        combinations[i] = to_fiat_tuple(combinations[i], geo_dim)

    return combinations


def get_data(ufl_element):
    "Get needed data to run tests."

    # Create fiat element.
    element = create_element(ufl_element)

    # The derivative order that we are interested in is the degree of
    # the element.
    if isinstance(element, FFCMixedElement):
        deriv_order = max([e.degree() for e in element.elements()])
    else:
        deriv_order = element.degree()

    # Get coordinates of the reference cell.
    cell = ufl_element.cell()
    ref_coords = reference_cell(cell.cellname()).get_vertices()

    # Get the locations of the fiat element dofs.
    elem_points = [list(L.pt_dict.keys())[0] for L in element.dual_basis()]

    # Add some random points.
    geo_dim = cell.geometric_dimension()
    points = elem_points + random_points[geo_dim]

    return (element, points, geo_dim, ref_coords, deriv_order)


def get_ffc_values(ufl_element):
    "Get the values from evaluate_basis and evaluate_basis_derivatives."

    # Get data from element and tabulate basis values
    element, points, geo_dim, ref_coords, deriv_order = get_data(ufl_element)

    # Get relevant element name
    element_name = get_element_name(ufl_element)

    # Create g++ code and compile
    num_coords = len(ref_coords)
    options = {"element": element_name,
               "dim": geo_dim,
               "num_points": len(points),
               "points": matrix(points),
               "cell_ref_coords": "double cell_ref_coords[{}][{}] = {}".format(num_coords, geo_dim, matrix(ref_coords)),
               "num_coords": num_coords}
    compile_gcc_code(ufl_element, evaluate_basis_code_fiat % options)

    # Loop over derivative order and compute values
    ffc_values = {}
    for n in range(deriv_order + 1):
        values = run_code(ufl_element, n)
        ffc_values[n] = values

    return ffc_values


def get_fiat_values(ufl_element):
    """Create a FIAT element and use it to tabulate the values on the
    reference element. The return values is a dictionary with keys
    equal to the derivative order and values is a matrix where each
    row is the basis values at a point.  E.g., {0:[[1,0,0],[0,1,0],
    [0,0,1]]}.

    """

    # Get data and tabulate basis values.
    element, points, geo_dim, ref_coords, deriv_order = get_data(ufl_element)
    values = element.tabulate(deriv_order, points)
    return_values = {}
    value_shape = element.value_shape()

    # Rearrange values to match what we get from evaluate_basis*()
    for n in range(deriv_order + 1):
        combinations = compute_derivative_combinations(n, geo_dim)
        vals = []
        # If we're evaluating the basis functions, use all points, but
        # if we're evaluating the derivatives, just use the 3
        # arbitrary points to avoid the number of tests exploding with
        # spacedim**2.
        if n == 0:
            new_points = points
        else:
            new_points = points[-3:]
        for p, point in enumerate(new_points):
            if n != 0:
                p += element.space_dimension()
            row = [[] for i in range(element.space_dimension())]
            for i in range(element.space_dimension()):
                if value_shape == ():
                    for deriv in combinations:
                        deriv_vals = values[deriv]
                        row[i].append(deriv_vals[i][p])
                elif len(value_shape) == 1:
                    for c in range(element.value_shape()[0]):
                        for deriv in combinations:
                            deriv_vals = values[deriv]
                            row[i].append(deriv_vals[i][c][p])
                elif len(value_shape) == 2:
                    for j in range(element.value_shape()[0]):
                        for k in range(element.value_shape()[1]):
                            for deriv in combinations:
                                deriv_vals = values[deriv]
                                row[i].append(deriv_vals[i][j][k][p])
                else:
                    print(values)
                    error("Did not expect tensor elements of rank > 2")
            new_row = []
            for r in row:
                new_row += r
            vals.append(new_row)
        return_values[n] = numpy.array(vals)
    return return_values


def verify_values(ufl_element, ref_values, ffc_values):
    "Check the values from evaluate_basis*() against some reference values."

    num_tests = len(ffc_values)
    if num_tests != len(ref_values):
        raise RuntimeError("The number of computed values is not equal to the number of reference values.")

    for deriv_order in range(num_tests):
        s = "evaluate_basis_derivatives, order = %d" % deriv_order
        e = abs(ffc_values[deriv_order] - ref_values[deriv_order])
        error = e.max()

        # Check that error is below critical tolerance
        assert not error > crit_tol, "{} failed, error={} (crit_tol: {})".format(s, error, crit_tol)

        # Print message if error is greater than tolerance
        if error > tol:
            print("{} ok: error={} (tol: {})".format(s, error, tol))

    return num_tests


# Test over different element types
@pytest.mark.parametrize("ufl_element", all_elements)
def test_element(ufl_element):
    "Test FFC elements against FIAT"

    print("\nVerifying element: {}".format(str(ufl_element)))

    # Generate element
    generate_element(ufl_element)

    # Get FFC values
    t = time.time()
    ffc_values = get_ffc_values(ufl_element)
    assert ffc_values is not None
    # debug("  time to compute FFC values: %f" % (time.time() - t))

    # Get FIAT values that are formatted in the same way as the values
    # from evaluate_basis and evaluate_basis_derivatives.
    fiat_values = get_fiat_values(ufl_element)

    # Compare FIAT and FFC values
    verify_values(ufl_element, fiat_values, ffc_values)



# -*- coding: utf-8 -*-
# Code generation format strings for UFC (Unified Form-assembly Code) version 2017.1.0
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2006-2017.

dofmap_combined = """
class %(classname)s: public ufc::dofmap
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::dofmap()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const char * signature() const final override
  {
%(signature)s
  }

  bool needs_mesh_entities(std::size_t d) const final override
  {
%(needs_mesh_entities)s
  }

  std::size_t topological_dimension() const final override
  {
%(topological_dimension)s
  }

  std::size_t global_dimension(const std::vector<std::size_t>&
                               num_global_entities) const final override
  {
%(global_dimension)s
  }

  std::size_t num_element_dofs() const final override
  {
%(num_element_dofs)s
  }

  std::size_t num_facet_dofs() const final override
  {
%(num_facet_dofs)s
  }

  std::size_t num_entity_dofs(std::size_t d) const final override
  {
%(num_entity_dofs)s
  }

  std::size_t num_entity_closure_dofs(std::size_t d) const final override
  {
%(num_entity_closure_dofs)s
  }

  void tabulate_dofs(std::size_t * dofs,
                     const std::vector<std::size_t>& num_global_entities,
                     const std::vector<std::vector<std::size_t>>& entity_indices) const final override
  {
%(tabulate_dofs)s
  }

  void tabulate_facet_dofs(std::size_t * dofs,
                           std::size_t facet) const final override
  {
%(tabulate_facet_dofs)s
  }

  void tabulate_entity_dofs(std::size_t * dofs,
                            std::size_t d, std::size_t i) const final override
  {
%(tabulate_entity_dofs)s
  }

  void tabulate_entity_closure_dofs(std::size_t * dofs,
                                    std::size_t d, std::size_t i) const final override
  {
%(tabulate_entity_closure_dofs)s
  }


  std::size_t num_sub_dofmaps() const final override
  {
%(num_sub_dofmaps)s
  }

  ufc::dofmap * create_sub_dofmap(std::size_t i) const final override
  {
%(create_sub_dofmap)s
  }

  ufc::dofmap * create() const final override
  {
%(create)s
  }

};
"""

dofmap_header = """
class %(classname)s: public ufc::dofmap
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const char * signature() const final override;

  bool needs_mesh_entities(std::size_t d) const final override;

  std::size_t topological_dimension() const final override;

  std::size_t global_dimension(const std::vector<std::size_t>&
                               num_global_entities) const final override;

  std::size_t num_element_dofs() const final override;

  std::size_t num_facet_dofs() const final override;

  std::size_t num_entity_dofs(std::size_t d) const final override;

  std::size_t num_entity_closure_dofs(std::size_t d) const final override;

  void tabulate_dofs(std::size_t * dofs,
                     const std::vector<std::size_t>& num_global_entities,
                     const std::vector<std::vector<std::size_t>>& entity_indices) const final override;

  void tabulate_facet_dofs(std::size_t * dofs,
                           std::size_t facet) const final override;

  void tabulate_entity_dofs(std::size_t * dofs,
                            std::size_t d, std::size_t i) const final override;

  void tabulate_entity_closure_dofs(std::size_t * dofs,
                            std::size_t d, std::size_t i) const final override;

  std::size_t num_sub_dofmaps() const final override;

  ufc::dofmap * create_sub_dofmap(std::size_t i) const final override;

  ufc::dofmap * create() const final override;

};
"""

dofmap_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::dofmap()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const char * %(classname)s::signature() const
{
%(signature)s
}

bool %(classname)s::needs_mesh_entities(std::size_t d) const
{
%(needs_mesh_entities)s
}

std::size_t %(classname)s::topological_dimension() const
{
%(topological_dimension)s
}

std::size_t %(classname)s::global_dimension(const std::vector<std::size_t>&
                                            num_global_entities) const
{
%(global_dimension)s
}

std::size_t %(classname)s::num_element_dofs() const
{
%(num_element_dofs)s
}

std::size_t %(classname)s::num_facet_dofs() const
{
%(num_facet_dofs)s
}

std::size_t %(classname)s::num_entity_dofs(std::size_t d) const
{
%(num_entity_dofs)s
}

std::size_t %(classname)s::num_entity_closure_dofs(std::size_t d) const
{
%(num_entity_closure_dofs)s
}

void %(classname)s::tabulate_dofs(std::size_t * dofs,
                                  const std::vector<std::size_t>& num_global_entities,
                                  const std::vector<std::vector<std::size_t>>& entity_indices) const
{
%(tabulate_dofs)s
}

void %(classname)s::tabulate_facet_dofs(std::size_t * dofs,
                                        std::size_t facet) const
{
%(tabulate_facet_dofs)s
}

void %(classname)s::tabulate_entity_dofs(std::size_t * dofs,
                                         std::size_t d, std::size_t i) const
{
%(tabulate_entity_dofs)s
}

void %(classname)s::tabulate_entity_closure_dofs(std::size_t * dofs,
                                             std::size_t d, std::size_t i) const
{
%(tabulate_entity_closure_dofs)s
}

std::size_t %(classname)s::num_sub_dofmaps() const
{
%(num_sub_dofmaps)s
}

ufc::dofmap * %(classname)s::create_sub_dofmap(std::size_t i) const
{
%(create_sub_dofmap)s
}

ufc::dofmap * %(classname)s::create() const
{
%(create)s
}
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This script is the command-line interface to FFC.

It parses command-line arguments and generates code from input UFL form files.
"""

# Copyright (C) 2004-2016 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Johan Jansson, 2005.
# Modified by Ola Skavhaug, 2006.
# Modified by Dag Lindbo, 2008.
# Modified by Kristian B. Oelgaard 2010.
# Modified by Martin Sandve Alnæs 2017.

from __future__ import print_function

# Python modules.
import sys
import getopt
import cProfile
import re
import string
import os
from os import curdir
from os import path
from os import getcwd

# UFL modules.
from ufl.log import UFLException
from ufl.algorithms import load_ufl_file
import ufl

# FFC modules.
from ffc.log import push_level, pop_level, set_indent, ffc_logger
from ffc.log import DEBUG, INFO, WARNING, ERROR
from ffc.parameters import default_parameters
from ffc import __version__ as FFC_VERSION, get_ufc_signature
from ffc.backends.ufc import __version__ as UFC_VERSION
from ffc.backends.ufc import get_include_path
from ffc.compiler import compile_form, compile_element
from ffc.formatting import write_code
from ffc.errorcontrol import compile_with_error_control


def print_error(msg):
    "Print error message (cannot use log system at top level)."
    print("\n".join(["*** FFC: " + line for line in msg.split("\n")]))


def info_version():
    "Print version number."
    print("""\
This is FFC, the FEniCS Form Compiler, version {0}.
UFC backend version {1}, signature {2}.
For further information, visit https://bitbucket.org/fenics-project/ffc/.

Python {3} on {4}
""".format(FFC_VERSION, UFC_VERSION, get_ufc_signature(),
           sys.version, sys.platform))


def info_usage():
    "Print usage information."
    info_version()
    print("""Usage: ffc [OPTION]... input.form

For information about the FFC command-line interface, refer to
the FFC man page which may invoked by 'man ffc' (if installed).
""")


def compile_ufl_data(ufd, prefix, parameters):
    if parameters["error_control"]:
        code_h, code_c = compile_with_error_control(ufd.forms,
                                                    ufd.object_names,
                                                    ufd.reserved_objects,
                                                    prefix,
                                                    parameters)
    elif len(ufd.forms) > 0:
        code_h, code_c = compile_form(ufd.forms, ufd.object_names,
                                      prefix=prefix,
                                      parameters=parameters)
    else:
        code_h, code_c = compile_element(ufd.elements, prefix=prefix,
                                         parameters=parameters)
    return code_h, code_c


def main(args=None):
    """This is the commandline tool for the python module ffc."""
    if args is None:
        args = sys.argv[1:]

    # Get command-line arguments
    try:
        if "-O" in args:
            args[args.index("-O")] = "-O2"
        opts, args = getopt.getopt(args, "hIVSdvsl:r:f:O:o:q:ep",
                                   ["help", "includes", "version", "signature", "debug", "verbose", "silent",
                                    "language=", "representation=", "optimize=",
                                    "output-directory=", "quadrature-rule=", "error-control",
                                    "profile"])
    except getopt.GetoptError:
        info_usage()
        print_error("Illegal command-line arguments.")
        return 1

    # Check for --help
    if ("-h", "") in opts or ("--help", "") in opts:
        info_usage()
        return 0

    # Check for --includes
    if ("-I", "") in opts or ("--includes", "") in opts:
        print(get_include_path())
        return 0

    # Check for --version
    if ("-V", "") in opts or ("--version", "") in opts:
        info_version()
        return 0

    # Check for --signature
    if ("-S", "") in opts or ("--signature", "") in opts:
        print(get_ufc_signature())
        return 0

    # Check that we get at least one file
    if len(args) == 0:
        print_error("Missing file.")
        return 1

    # Get parameters
    parameters = default_parameters()

    # Choose WARNING as default for script
    parameters["log_level"] = WARNING

    # Set default value (not part of in parameters[])
    enable_profile = False

    # Parse command-line parameters
    for opt, arg in opts:
        if opt in ("-v", "--verbose"):
            parameters["log_level"] = INFO
        elif opt in ("-d", "--debug"):
            parameters["log_level"] = DEBUG
        elif opt in ("-s", "--silent"):
            parameters["log_level"] = ERROR
        elif opt in ("-l", "--language"):
            parameters["format"] = arg
        elif opt in ("-r", "--representation"):
            parameters["representation"] = arg
        elif opt in ("-q", "--quadrature-rule"):
            parameters["quadrature_rule"] = arg
        elif opt == "-f":
            if len(arg.split("=")) == 2:
                (key, value) = arg.split("=")
                default = parameters.get(key)
                if isinstance(default, int):
                    value = int(value)
                elif isinstance(default, float):
                    value = float(value)
                parameters[key] = value
            elif len(arg.split("==")) == 1:
                key = arg.split("=")[0]
                if key.startswith("no-"):
                    key = key[3:]
                    value = False
                else:
                    value = True
                parameters[key] = value
            else:
                info_usage()
                return 1
        elif opt in ("-O", "--optimize"):
            parameters["optimize"] = bool(int(arg))
        elif opt in ("-o", "--output-directory"):
            parameters["output_dir"] = arg
        elif opt in ("-e", "--error-control"):
            parameters["error_control"] = True
        elif opt in ("-p", "--profile"):
            enable_profile = True

    # Set log_level
    push_level(parameters["log_level"])

    # FIXME: This is terrible!
    # Set UFL precision
    #ufl.constantvalue.precision = int(parameters["precision"])

    # Print a versioning message if verbose output was requested
    if parameters["log_level"] <= INFO:
        info_version()

    # Call parser and compiler for each file
    resultcode = 0
    init_indent = ffc_logger._indent_level
    try:
        resultcode = _compile_files(args, parameters, enable_profile)
    finally:
        # Reset logging level and indent
        pop_level()
        set_indent(init_indent)

    return resultcode


def _compile_files(args, parameters, enable_profile):
    # Call parser and compiler for each file
    for filename in args:

        # Get filename prefix and suffix
        prefix, suffix = os.path.splitext(os.path.basename(filename))
        suffix = suffix.replace(os.path.extsep, "")

        # Check file suffix
        if suffix != "ufl":
            print_error("Expecting a UFL form file (.ufl).")
            return 1

        # Remove weird characters (file system allows more than the C
        # preprocessor)
        prefix = re.subn("[^{}]".format(string.ascii_letters + string.digits + "_"), "!", prefix)[0]
        prefix = re.subn("!+", "_", prefix)[0]

        # Turn on profiling
        if enable_profile:
            pr = cProfile.Profile()
            pr.enable()

        # Load UFL file
        ufd = load_ufl_file(filename)

        # Previously wrapped in try-except, disabled to actually get information we need
        #try:

        # Generate code
        code_h, code_c = compile_ufl_data(ufd, prefix, parameters)

        # Write to file
        write_code(code_h, code_c, prefix, parameters)

        #except Exception as exception:
        #    # Catch exceptions only when not in debug mode
        #    if parameters["log_level"] <= DEBUG:
        #        raise
        #    else:
        #        print("")
        #        print_error(str(exception))
        #        print_error("To get more information about this error, rerun FFC with --debug.")
        #        return 1

        # Turn off profiling and write status to file
        if enable_profile:
            pr.disable()
            pfn = "ffc_{0}.profile".format(prefix)
            pr.dump_stats(pfn)
            print("Wrote profiling info to file {0}".format(pfn))

    return 0



# Copyright (C) 2016 Jan Blechta
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

def git_commit_hash():
    """Return git changeset hash (returns "unknown" if changeset is
    not known)
    """
    return "b'de87b121569a0046f3011563e777176688eab916'"



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

class PRECEDENCE:
    "An enum-like class for C operator precedence levels."
    HIGHEST = 0
    LITERAL = 0
    SYMBOL = 0

    #SCOPE = 1

    POST_INC = 2
    POST_DEC = 2
    CALL = 2
    SUBSCRIPT = 2
    #MEMBER = 2
    #PTR_MEMBER = 2

    PRE_INC = 3
    PRE_DEC = 3
    NOT = 3
    BIT_NOT = 3
    POS = 3
    NEG = 3
    DEREFERENCE = 3
    ADDRESSOF = 3
    SIZEOF = 3

    MUL = 4
    DIV = 4
    MOD = 4

    ADD = 5
    SUB = 5

    LT = 6
    LE = 6
    GT = 6
    GE = 6

    EQ = 7
    NE = 7

    BIT_AND = 8
    BIT_XOR = 9
    BIT_OR = 10
    AND = 11
    OR = 12

    CONDITIONAL = 13
    ASSIGN = 13

    #COMMA = 14

    LOWEST = 15



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>

"""FFC/UFC specific symbol naming."""


from ffc.log import error


# TODO: Move somewhere else
def num_coordinate_component_dofs(coordinate_element):
    """Get the number of dofs for a coordinate component for this degree.

    This is a local hack that works for Lagrange 1-3, better
    would be to get this passed by ffc from fiat through the ir.
    The table data is to messy to figure out a clean design for that quickly.
    """
    from ufl.cell import num_cell_entities
    degree = coordinate_element.degree()
    cell = coordinate_element.cell()
    tdim = cell.topological_dimension()
    cellname = cell.cellname()
    d = 0
    for entity_dim in range(tdim+1):
        # n = dofs per cell entity
        if entity_dim == 0:
            n = 1
        elif entity_dim == 1:
            n = degree - 1
        elif entity_dim == 2:
            n = (degree - 2)*(degree - 1) // 2
        elif entity_dim == 3:
            n = (degree - 3)*(degree - 2)*(degree - 1) // 6
        else:
            error("Entity dimension out of range")
        # Accumulate
        num_entities = num_cell_entities[cellname][entity_dim]
        d += num_entities * n
    return d


# TODO: Get restriction postfix from somewhere central
def ufc_restriction_offset(restriction, length):
    if restriction == "-":
        return length
    else:
        return 0



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

"""Assigning symbols to computational graph nodes."""

import numpy
from ufl import product

from ffc.uflacs.analysis.crsarray import CRSArray
from ffc.uflacs.analysis.valuenumbering import ValueNumberer
from ffc.uflacs.analysis.expr_shapes import total_shape


def build_node_shapes(V):
    """Build total shapes for each node in list representation of expression graph.

    V is an array of ufl expressions, possibly nonscalar and with free indices.

    Returning a CRSArray where row i is the total shape of V[i].
    """
    # Dimensions of returned CRSArray
    nv = len(V)
    k = 0

    # Store shapes intermediately in an array of tuples
    V_shapes = numpy.empty(nv, dtype=object)
    for i, v in enumerate(V):
        # Compute total shape of V[i]
        tsh = total_shape(v)
        V_shapes[i] = tsh
        # Count number of elements for CRSArray representation
        k += len(tsh)

    # Return a more memory efficient CRSArray representation
    return CRSArray.from_rows(V_shapes, nv, k, int)


def build_node_sizes(V_shapes):
    "Compute all the products of a sequence of shapes."
    nv = len(V_shapes)
    V_sizes = numpy.zeros(nv, dtype=int)
    for i, sh in enumerate(V_shapes):
        V_sizes[i] = product(sh)
    return V_sizes


def build_node_symbols(V, e2i, V_shapes, V_sizes):
    """Tabulate scalar value numbering of all nodes in a a list based representation of an expression graph.

    Returns:
    V_symbols - CRSArray of symbols (value numbers) of each component of each node in V.
    total_unique_symbols - The number of symbol values assigned to unique scalar components of the nodes in V.
    """
    # "Sparse" int matrix for storing variable number of entries (symbols) per row (vertex),
    # with a capasity bounded by the number of scalar subexpressions including repetitions
    V_symbols = CRSArray(len(V), sum(V_sizes), int)

    # Visit each node with value numberer algorithm, storing the result for each as a row in the V_symbols CRSArray
    value_numberer = ValueNumberer(e2i, V_sizes, V_symbols)
    for i, v in enumerate(V):
        V_symbols.push_row(value_numberer(v, i))

    # Get the actual number of symbols created
    total_unique_symbols = value_numberer.symbol_count

    # assert all(x < total_unique_symbols for x in V_symbols.data)
    # assert (total_unique_symbols-1) in V_symbols.data

    return V_symbols, total_unique_symbols


def build_graph_symbols(V, e2i, DEBUG):
    """Tabulate scalar value numbering of all nodes in a a list based representation of an expression graph.

    Returns:
    V_shapes - CRSArray of the total shapes of nodes in V.
    V_symbols - CRSArray of symbols (value numbers) of each component of each node in V.
    total_unique_symbols - The number of symbol values assigned to unique scalar components of the nodes in V.
    """
    # Compute the total shape (value shape x index dimensions) for each node
    V_shapes = build_node_shapes(V)

    # Compute the total value size for each node
    V_sizes = build_node_sizes(V_shapes)

    # Mark values with symbols
    V_symbols, total_unique_symbols = build_node_symbols(V, e2i, V_shapes, V_sizes)

    return V_shapes, V_symbols, total_unique_symbols



# -*- coding: utf-8 -*-
# Code generation format strings for UFC (Unified Form-assembly Code) version 2017.1.0
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2006-2017.

finite_element_combined = """
class %(classname)s: public ufc::finite_element
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::finite_element()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const char * signature() const final override
  {
%(signature)s
  }

  ufc::shape cell_shape() const final override
  {
%(cell_shape)s
  }

  std::size_t topological_dimension() const final override
  {
%(topological_dimension)s
  }

  std::size_t geometric_dimension() const final override
  {
%(geometric_dimension)s
  }

  std::size_t space_dimension() const final override
  {
%(space_dimension)s
  }

  std::size_t value_rank() const final override
  {
%(value_rank)s
  }

  std::size_t value_dimension(std::size_t i) const final override
  {
%(value_dimension)s
  }

  std::size_t value_size() const final override
  {
%(value_size)s
  }

  std::size_t reference_value_rank() const final override
  {
%(reference_value_rank)s
  }

  std::size_t reference_value_dimension(std::size_t i) const final override
  {
%(reference_value_dimension)s
  }

  std::size_t reference_value_size() const final override
  {
%(reference_value_size)s
  }

  std::size_t degree() const final override
  {
%(degree)s
  }

  const char * family() const final override
  {
%(family)s
  }

  static void _evaluate_basis(std::size_t i,
                              double * values,
                              const double * x,
                              const double * coordinate_dofs,
                              int cell_orientation)
  {
%(evaluate_basis)s
  }

  void evaluate_basis(std::size_t i,
                      double * values,
                      const double * x,
                      const double * coordinate_dofs,
                      int cell_orientation) const final override
  {
    _evaluate_basis(i, values, x, coordinate_dofs, cell_orientation);
  }

  static void _evaluate_basis_all(double * values,
                                  const double * x,
                                  const double * coordinate_dofs,
                                  int cell_orientation)
  {
%(evaluate_basis_all)s
  }

  void evaluate_basis_all(double * values,
                          const double * x,
                          const double * coordinate_dofs,
                          int cell_orientation) const final override
  {
    _evaluate_basis_all(values, x, coordinate_dofs, cell_orientation);
  }

  static void _evaluate_basis_derivatives(std::size_t i,
                                          std::size_t n,
                                          double * values,
                                          const double * x,
                                          const double * coordinate_dofs,
                                          int cell_orientation)
  {
%(evaluate_basis_derivatives)s
  }

  void evaluate_basis_derivatives(std::size_t i,
                                  std::size_t n,
                                  double * values,
                                  const double * x,
                                  const double * coordinate_dofs,
                                  int cell_orientation) const final override
  {
    _evaluate_basis_derivatives(i, n, values, x, coordinate_dofs, cell_orientation);
  }

  static void _evaluate_basis_derivatives_all(std::size_t n,
                                              double * values,
                                              const double * x,
                                              const double * coordinate_dofs,
                                              int cell_orientation)
  {
%(evaluate_basis_derivatives_all)s
  }

  void evaluate_basis_derivatives_all(std::size_t n,
                                      double * values,
                                      const double * x,
                                      const double * coordinate_dofs,
                                      int cell_orientation) const final override
  {
    _evaluate_basis_derivatives_all(n, values, x, coordinate_dofs, cell_orientation);
  }

  double evaluate_dof(std::size_t i,
                      const ufc::function& f,
                      const double * coordinate_dofs,
                      int cell_orientation,
                      const ufc::cell& c) const final override
  {
%(evaluate_dof)s
  }

  void evaluate_dofs(double * values,
                             const ufc::function& f,
                             const double * coordinate_dofs,
                             int cell_orientation,
                             const ufc::cell& c) const final override
  {
%(evaluate_dofs)s
  }

  void interpolate_vertex_values(double * vertex_values,
                                 const double * dof_values,
                                 const double * coordinate_dofs,
                                 int cell_orientation,
                                 const ufc::cell& c) const final override
  {
%(interpolate_vertex_values)s
  }

  void tabulate_dof_coordinates(double * dof_coordinates,
                                const double * coordinate_dofs) const final override
  {
%(tabulate_dof_coordinates)s
  }

  std::size_t num_sub_elements() const final override
  {
%(num_sub_elements)s
  }

  ufc::finite_element * create_sub_element(std::size_t i) const final override
  {
%(create_sub_element)s
  }

  ufc::finite_element * create() const final override
  {
%(create)s
  }

};
"""

finite_element_header = """
class %(classname)s: public ufc::finite_element
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const char * signature() const final override;

  ufc::shape cell_shape() const final override;

  std::size_t topological_dimension() const final override;

  std::size_t geometric_dimension() const final override;

  std::size_t space_dimension() const final override;

  std::size_t value_rank() const final override;

  std::size_t value_dimension(std::size_t i) const final override;

  std::size_t value_size() const final override;

  std::size_t reference_value_rank() const final override;

  std::size_t reference_value_dimension(std::size_t i) const final override;

  std::size_t reference_value_size() const final override;

  std::size_t degree() const final override;

  const char * family() const final override;

  static void _evaluate_basis(std::size_t i,
                              double * values,
                              const double * x,
                              const double * coordinate_dofs,
                              int cell_orientation);

  void evaluate_basis(std::size_t i,
                      double * values,
                      const double * x,
                      const double * coordinate_dofs,
                      int cell_orientation) const final override
  {
    _evaluate_basis(i, values, x, coordinate_dofs, cell_orientation);
  }

  static void _evaluate_basis_all(double * values,
                                  const double * x,
                                  const double * coordinate_dofs,
                                  int cell_orientation);

  void evaluate_basis_all(double * values,
                          const double * x,
                          const double * coordinate_dofs,
                          int cell_orientation) const final override
  {
    _evaluate_basis_all(values, x, coordinate_dofs, cell_orientation);
  }

  static void _evaluate_basis_derivatives(std::size_t i,
                                          std::size_t n,
                                          double * values,
                                          const double * x,
                                          const double * coordinate_dofs,
                                          int cell_orientation);

  void evaluate_basis_derivatives(std::size_t i,
                                  std::size_t n,
                                  double * values,
                                  const double * x,
                                  const double * coordinate_dofs,
                                  int cell_orientation) const final override
  {
    _evaluate_basis_derivatives(i, n, values, x, coordinate_dofs, cell_orientation);
  }

  static void _evaluate_basis_derivatives_all(std::size_t n,
                                              double * values,
                                              const double * x,
                                              const double * coordinate_dofs,
                                              int cell_orientation);

  void evaluate_basis_derivatives_all(std::size_t n,
                                      double * values,
                                      const double * x,
                                      const double * coordinate_dofs,
                                      int cell_orientation) const final override
  {
    _evaluate_basis_derivatives_all(n, values, x, coordinate_dofs, cell_orientation);
  }

  double evaluate_dof(std::size_t i,
                      const ufc::function& f,
                      const double * coordinate_dofs,
                      int cell_orientation,
                      const ufc::cell& c) const final override;

  void evaluate_dofs(double * values,
                     const ufc::function& f,
                     const double * coordinate_dofs,
                     int cell_orientation,
                     const ufc::cell& c) const final override;

  void interpolate_vertex_values(double * vertex_values,
                                 const double * dof_values,
                                 const double * coordinate_dofs,
                                 int cell_orientation,
                                 const ufc::cell& c) const final override;

  void tabulate_dof_coordinates(double * coordinates,
                                const double * coordinate_dofs) const final override;

  std::size_t num_sub_elements() const final override;

  ufc::finite_element * create_sub_element(std::size_t i) const final override;

  ufc::finite_element * create() const final override;

};
"""

finite_element_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::finite_element()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s()
{
%(destructor)s
}

const char * %(classname)s::signature() const
{
%(signature)s
}

ufc::shape %(classname)s::cell_shape() const
{
%(cell_shape)s
}

std::size_t %(classname)s::topological_dimension() const
{
%(topological_dimension)s
}

std::size_t %(classname)s::geometric_dimension() const
{
%(geometric_dimension)s
}

std::size_t %(classname)s::space_dimension() const
{
%(space_dimension)s
}

std::size_t %(classname)s::value_rank() const
{
%(value_rank)s
}

std::size_t %(classname)s::value_dimension(std::size_t i) const
{
%(value_dimension)s
}

std::size_t %(classname)s::value_size() const
{
%(value_size)s
}

std::size_t %(classname)s::reference_value_rank() const
{
%(reference_value_rank)s
}

std::size_t %(classname)s::reference_value_dimension(std::size_t i) const
{
%(reference_value_dimension)s
}

std::size_t %(classname)s::reference_value_size() const
{
%(reference_value_size)s
}

std::size_t %(classname)s::degree() const
{
%(degree)s
}

const char * %(classname)s::family() const
{
%(family)s
}

void %(classname)s::_evaluate_basis(std::size_t i,
                                    double * values,
                                    const double * x,
                                    const double * coordinate_dofs,
                                    int cell_orientation)
{
%(evaluate_basis)s
}

void %(classname)s::_evaluate_basis_all(double * values,
                                        const double * x,
                                        const double * coordinate_dofs,
                                        int cell_orientation)
{
%(evaluate_basis_all)s
}

void %(classname)s::_evaluate_basis_derivatives(std::size_t i,
                                                std::size_t n,
                                                double * values,
                                                const double * x,
                                                const double * coordinate_dofs,
                                                int cell_orientation)
{
%(evaluate_basis_derivatives)s
}

void %(classname)s::_evaluate_basis_derivatives_all(std::size_t n,
                                                    double * values,
                                                    const double * x,
                                                    const double * coordinate_dofs,
                                                    int cell_orientation)
{
%(evaluate_basis_derivatives_all)s
}

double %(classname)s::evaluate_dof(std::size_t i,
                                   const ufc::function& f,
                                   const double * coordinate_dofs,
                                   int cell_orientation,
                                   const ufc::cell& c) const
{
%(evaluate_dof)s
}

void %(classname)s::evaluate_dofs(double * values,
                                  const ufc::function& f,
                                  const double * coordinate_dofs,
                                  int cell_orientation,
                                  const ufc::cell& c) const
{
%(evaluate_dofs)s
}

void %(classname)s::interpolate_vertex_values(double * vertex_values,
                                              const double * dof_values,
                                              const double * coordinate_dofs,
                                              int cell_orientation,
                                              const ufc::cell& c) const
{
%(interpolate_vertex_values)s
}

void %(classname)s::tabulate_dof_coordinates(double * dof_coordinates,
                                             const double * coordinate_dofs) const
{
%(tabulate_dof_coordinates)s
}

std::size_t %(classname)s::num_sub_elements() const
{
%(num_sub_elements)s
}

ufc::finite_element * %(classname)s::create_sub_element(std::size_t i) const
{
%(create_sub_element)s
}

ufc::finite_element * %(classname)s::create() const
{
%(create)s
}
"""



#!/usr/bin/env py.test
# -*- coding: utf-8 -*-

"""
Tests generating code for the different ufc integral types,
checking that the variations between them are correctly
reflected in the code.
"""

from ufl import *
#from ffc.uflacs.backends.? import ?

def test_cell_integral_body(gtest):
    """
      class cell_integral
        virtual void tabulate_tensor(double* A,
                                     const double * const * w,
                                     const double* coordinate_dofs,
                                     int cell_orientation) const = 0;
    """

    pre = """
    """

    post = """
    """

    code = """
    """

    gtest.add(pre + code + post)


def test_exterior_facet_integral_body(gtest):
    """
      class exterior_facet_integral
        virtual void tabulate_tensor(double* A,
                                     const double * const * w,
                                     const double* coordinate_dofs,
                                     std::size_t facet) const = 0;
    """

    pre = """
    """

    post = """
    """

    code = """
    """

    gtest.add(pre + code + post)


def test_interior_facet_integral_body(gtest):
    """
      class interior_facet_integral
        virtual void tabulate_tensor(double* A,
                                     const double * const * w,
                                     const double* coordinate_dofs_0,
                                     const double* coordinate_dofs_1,
                                     std::size_t facet_0,
                                     std::size_t facet_1) const = 0;
    """

    pre = """
    """

    post = """
    """

    code = """
    """

    gtest.add(pre + code + post)


def test_vertex_integral_body(gtest):
    """
      class vertex_integral
        virtual void tabulate_tensor(double* A,
                                     const double * const * w,
                                     const double* coordinate_dofs,
                                     std::size_t vertex) const = 0;
    """

    pre = """
    """

    post = """
    """

    code = """
    """

    gtest.add(pre + code + post)


def test_custom_integral_body(gtest):
    """
      class custom_integral
        virtual void tabulate_tensor(double* A,
                                     const double * const * w,
                                     const double* coordinate_dofs,
                                     std::size_t num_quadrature_points,
                                     const double* quadrature_points,
                                     const double* quadrature_weights) const = 0;
    """

    pre = """
    """

    post = """
    """

    code = """
    """

    gtest.add(pre + code + post)



# -*- coding: utf-8 -*-
# Copyright (C) 2011-2016 Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

"""Tools for indentation-aware code string stitching.

When formatting an AST into a string, it's better to collect
lists of snippets and then join them than adding the pieces
continually, which gives O(n^2) behaviour w.r.t. AST size n.
"""

from six import string_types

class Indented(object):
    """Class to mark a collection of snippets for indentation.

    This way nested indentations can be handled by adding the prefix
    spaces only once to each line instead of splitting and indenting
    substrings repeatedly.
    """
    # Try to keep memory overhead low:
    __slots__ = ("body",)
    def __init__(self, body):
        # Body can be any valid snippet format
        self.body = body

def iter_indented_lines(snippets, level=0):
    """Iterate over indented string lines from a snippets data structure.

    The snippets object can be built recursively using the following types:

    - str: Split and yield as one line at a time indented to the appropriate level.

    - Indented: Yield the lines within this object indented by one level.

    - tuple,list: Yield lines from recursive application of this function to list items.

    """
    tabsize = 4
    indentation = ' ' * (tabsize * level)
    if isinstance(snippets, string_types):
        for line in snippets.split("\n"):
            yield indentation + line
    elif isinstance(snippets, Indented):
        for line in iter_indented_lines(snippets.body, level+1):
            yield line
    elif isinstance(snippets, (tuple, list)):
        for part in snippets:
            for line in iter_indented_lines(part, level):
                yield line
    else:
        raise RuntimeError("Unexpected type %s:\n%s" % (type(snippets), str(snippets)))

def format_indented_lines(snippets, level=0):
    "Format recursive sequences of indented lines as one string."
    return "\n".join(iter_indented_lines(snippets, level))



# -*- coding: utf-8 -*-
# Copyright (C) 2009-2016 Anders Logg and Martin Sandve Alnæs
#
# This file is part of UFLACS.
#
# UFLACS is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFLACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFLACS. If not, see <http://www.gnu.org/licenses/>.

# Note: Most of the code in this file is a direct translation from the old implementation in FFC


from ufl import product
from ffc.uflacs.backends.ufc.generator import ufc_generator
from ffc.uflacs.backends.ufc.utils import generate_return_new_switch


def affine_weights(dim): # FIXME: This is used where we still assume an affine mesh. Get rid of all places that use it.
    "Compute coefficents for mapping from reference to physical element"
    if dim == 1:
        return lambda X: (1.0 - X[0], X[0])
    elif dim == 2:
        return lambda X: (1.0 - X[0] - X[1], X[0], X[1])
    elif dim == 3:
        return lambda X: (1.0 - X[0] - X[1] - X[2], X[0], X[1], X[2])


class ufc_finite_element(ufc_generator):
    def __init__(self):
        ufc_generator.__init__(self, "finite_element")

    def topological_dimension(self, L, ir):
        tdim = ir["topological_dimension"]
        return L.Return(L.LiteralInt(tdim))

    def geometric_dimension(self, L, ir):
        gdim = ir["geometric_dimension"]
        return L.Return(L.LiteralInt(gdim))

    def degree(self, L, ir):
        degree = ir["degree"]
        return L.Return(L.LiteralInt(degree))

    def family(self, L, ir):
        family = ir["family"]
        return L.Return(L.LiteralString(family))

    def cell_shape(self, L, ir):
        name = ir["cell_shape"]
        return L.Return(L.Symbol(name))

    def space_dimension(self, L, ir):
        value = ir["space_dimension"]
        return L.Return(L.LiteralInt(value))

    def value_rank(self, L, ir):
        sh = ir["value_dimension"]
        return L.Return(L.LiteralInt(len(sh)))

    def value_size(self, L, ir):
        sh = ir["value_dimension"]
        return L.Return(L.LiteralInt(product(sh)))

    def value_dimension(self, L, ir):
        i = L.Symbol("i")
        sh = ir["value_dimension"]
        cases = [(L.LiteralInt(j), L.Return(L.LiteralInt(k))) for j, k in enumerate(sh)]
        default = L.Return(L.LiteralInt(0))
        return L.Switch(i, cases, default=default, autoscope=False, autobreak=False)

    def reference_value_rank(self, L, ir):
        sh = ir["reference_value_dimension"]
        return L.Return(L.LiteralInt(len(sh)))

    def reference_value_size(self, L, ir):
        sh = ir["reference_value_dimension"]
        return L.Return(L.LiteralInt(product(sh)))

    def reference_value_dimension(self, L, ir):
        i = L.Symbol("i")
        sh = ir["reference_value_dimension"]
        cases = [(L.LiteralInt(j), L.Return(L.LiteralInt(k))) for j, k in enumerate(sh)]
        default = L.Return(L.LiteralInt(0))
        return L.Switch(i, cases, default=default, autoscope=False, autobreak=False)

    def evaluate_reference_basis(self, L, ir): # FIXME: NEW implement!
        from ffc.uflacs.backends.ufc.evaluatebasis import generate_evaluate_reference_basis
        return generate_evaluate_reference_basis(ir["evaluate_reference_basis"])

    def evaluate_reference_basis_derivatives(self, L, ir): # FIXME: NEW implement!
        data = ir["evaluate_reference_basis_derivatives"]
        return "FIXME"

    def evaluate_basis(self, L, ir): # FIXME: Get rid of this
        data = ir["evaluate_basis"]
        return "FIXME"

    def evaluate_basis_derivatives(self, L, ir): # FIXME: Get rid of this
        # FIXME: port this, then translate into reference version
        data = ir["evaluate_basis_derivatives"]
        return "FIXME"

    def evaluate_dof(self, L, ir): # FIXME: Get rid of this
        # FIXME: port this, then translate into reference version
        data = ir["evaluate_dof"]
        return "FIXME"


    def evaluate_basis_all(self, L, ir):
        # FIXME: port this, then translate into reference version
        data = ir["evaluate_basis_all"]
        return "FIXME"

    def evaluate_basis_derivatives_all(self, L, ir):
        # FIXME: port this, then translate into reference version
        data = ir["evaluate_basis_derivatives_all"]
        return "FIXME"

    def evaluate_dofs(self, L, ir):
        """Generate code for evaluate_dofs."""
        # FIXME: port this, then translate into reference version
        """
        - evaluate_dof needs to be split into invert_mapping + evaluate_dof or similar?

          f = M fhat;  nu(f) = nu(M fhat) = nuhat(M^-1 f) = sum_i w_i M^-1 f(x_i)

          // Get fixed set of points on reference element
          num_points = element->num_dof_evaluation_points();
          double X[num_points*tdim];
          element->tabulate_dof_evaluation_points(X);

          // Compute geometry in these points
          domain->compute_geometry(reference_points, num_point, X, J, detJ, K, coordinate_dofs, cell_orientation);

          // Computed by dolfin
          for ip
            fvalues[ip][:] = f.evaluate(point[ip])[:];

          // Finally: nu_j(f) = sum_component sum_ip weights[j][ip][component] fvalues[ip][component]
          element->evaluate_dofs(fdofs, fvalues, J, detJ, K)
        """

        return "FIXME" + ir["evaluate_dofs"]

    def interpolate_vertex_values(self, L, ir): # FIXME: port this
        # FIXME: port this, then translate into reference version
        return "FIXME" + ir["interpolate_vertex_values"]

    def _tabulate_dof_reference_coordinates(self, L, ir):
        """TODO: Add this signature to finite_element:
        /// Tabulate the reference coordinates of all dofs on a cell
        virtual void tabulate_dof_reference_coordinates(double * X) const = 0;
        """
        pass

    def tabulate_dof_coordinates(self, L, ir): # FIXME: port this
        # TODO: For a transition period, let finite_element and dofmap depend on a class affine_<cellname>_domain?
        # TODO: Call _tabulate_dof_reference_coordinates to tabulate X[ndofs][tdim],
        # then call affine_domain::compute_physical_coordinates(x, X, coordinate_dofs)

        ir = ir["tabulate_dof_coordinates"]

        # Raise error if tabulate_dof_coordinates is ill-defined
        if not ir:
            msg = "tabulate_dof_coordinates is not defined for this element"
            return L.Comment(msg) #L.Raise(msg) # TODO: Error handling

        # Extract coordinates and cell dimension
        gdim = ir["gdim"]
        tdim = ir["tdim"]
        points = ir["points"]

        # Output argument
        dof_coordinates = L.FlattenedArray(L.Symbol("dof_coordinates"), dims=(len(points), gdim))

        # Input argument
        coordinate_dofs = L.Symbol("coordinate_dofs")

        # Reference coordinates
        dof_reference_coordinates = L.Symbol("dof_reference_coordinates")
        dof_reference_coordinate_values = [X[j] for X in points for j in range(tdim)]

        # Loop indices
        i = L.Symbol("i")
        k = L.Symbol("k")
        ip = L.Symbol("ip")

        # Basis symbol
        phi = L.Symbol("phi")

        # TODO: This code assumes an affine coordinate field.
        #       Ok for now in here, this function must be removed anyway.
        # Create code for evaluating affine coordinate basis functions
        num_scalar_xdofs = tdim + 1
        cg1_basis = affine_weights(tdim)
        phi_values = [phi_comp for X in points for phi_comp in cg1_basis(X)]
        assert len(phi_values) == len(points) * num_scalar_xdofs

        code = [
            L.ArrayDecl("static const double", dof_reference_coordinates,
                        (len(points) * tdim,),
                        values=dof_reference_coordinate_values),
            L.ArrayDecl("const double", phi,
                        (len(points) * num_scalar_xdofs,),
                        values=phi_values),
            L.ForRange(ip, 0, len(points), body=
                L.ForRange(i, 0, gdim, body=
                    L.ForRange(k, 0, num_scalar_xdofs, body=
                        L.AssignAdd(dof_coordinates[ip][i],
                                    coordinate_dofs[gdim*k + i]
                                    * phi[ip*num_scalar_xdofs + k])))),
            ]
        return L.StatementList(code)

    def num_sub_elements(self, L, ir):
        n = ir["num_sub_elements"]
        return L.Return(L.LiteralInt(n))

    def create_sub_element(self, L, ir):
        i = L.Symbol("i")
        classnames = ir["create_sub_element"]
        return generate_return_new_switch(L, i, classnames, factory=ir["jit"])



# -*- coding: utf-8 -*-
"Utility functions for quadrature representation."

# Copyright (C) 2007-2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2007-03-16
# Last changed: 2015-03-27
#
# Hacked by Marie E. Rognes 2013
# Modified by Anders Logg 2014
# Modified by Lizao Li 2015

# Python modules.
import numpy

# FFC modules.
from ffc.log import debug, error
from ffc.cpp import format


def create_psi_tables(tables, eliminate_zeros, entity_type):
    "Create names and maps for tables and non-zero entries if appropriate."

    debug("\nQG-utils, psi_tables:\n" + str(tables))

    # Create element map {points:{element:number,},} and a plain
    # dictionary {name:values,}.
    element_map, flat_tables = flatten_psi_tables(tables, entity_type)
    debug("\nQG-utils, psi_tables, flat_tables:\n" + str(flat_tables))

    # Reduce tables such that we only have those tables left with
    # unique values. Create a name map for those tables that are
    # redundant.
    name_map, unique_tables = unique_psi_tables(flat_tables, eliminate_zeros)
    debug("\nQG-utils, psi_tables, unique_tables:\n" + str(unique_tables))
    debug("\nQG-utils, psi_tables, name_map:\n" + str(name_map))

    return (element_map, name_map, unique_tables)


def flatten_psi_tables(tables, entity_type):
    """Create a 'flat' dictionary of tables with unique names and a name
    map that maps number of quadrature points and element name to a
    unique element number.

    Input tables on the format for scalar and non-scalar elements respectively:
      tables[num_points][element][entity][derivs][ip][dof]
      tables[num_points][element][entity][derivs][ip][component][dof]

    Planning to change this into:
      tables[num_points][element][avg][entity][derivs][ip][dof]
      tables[num_points][element][avg][entity][derivs][ip][component][dof]

    Returns:
      element_map - { num_quad_points: {ufl_element: element_number} }.
      flat_tables - { unique_table_name: values[ip,dof] }.

    """

    generate_psi_name = format["psi name"]

    def sorted_items(mapping, **sorted_args):
        return [(k, mapping[k]) for k in sorted(mapping.keys(), **sorted_args)]

    # Initialise return values and element counter.
    flat_tables = {}
    element_map = {}
    counter = 0

    # There's a separate set of tables for each number of quadrature points
    for num_points, element_tables in sorted_items(tables):
        element_map[num_points] = {}

        # There's a set of tables for each element
        for element, avg_tables in sorted_items(element_tables,
                                                key=lambda x: str(x)):
            element_map[num_points][element] = counter

            # There's a set of tables for non-averaged and averaged
            # (averaged only occurs with num_points == 1)
            for avg, entity_tables in sorted_items(avg_tables):

                # There's a set of tables for each entity number (only
                # 1 for the cell, >1 for facets and vertices)
                for entity, derivs_tables in sorted_items(entity_tables):

                    # There's a set of tables for each derivative combination
                    for derivs, fiat_tables in sorted_items(derivs_tables):
                        # Flatten fiat_table for tensor-valued basis
                        # This is necessary for basic
                        # (non-tensor-product) tensor elements

                        if len(numpy.shape(fiat_tables)) > 3:
                            shape = fiat_tables.shape
                            value_shape = shape[1:-1]
                            fiat_tables = fiat_tables.reshape((shape[0],
                                                               numpy.product(value_shape),
                                                               shape[-1]))

                        # Transform fiat_tables to a list of tables on
                        # the form psi_table[dof][ip] for each scalar
                        # component
                        if element.value_shape():
                            # Table is on the form
                            # fiat_tables[ip][component][dof].
                            transposed_table = numpy.transpose(fiat_tables, (1, 2, 0))
                            component_tables = list(enumerate(transposed_table))
                        else:
                            # Scalar element, table is on the form
                            # fiat_tables[ip][dof].  Using () for the
                            # component because generate_psi_name
                            # expects that
                            component_tables = [((),
                                                 numpy.transpose(fiat_tables))]

                        # Iterate over the innermost tables for each
                        # scalar component
                        for component, psi_table in component_tables:

                            # Generate the table name.
                            name = generate_psi_name(counter, entity_type,
                                                     entity, component,
                                                     derivs, avg)

                            # Verify uniqueness of names
                            if name in flat_tables:
                                error("Table name is not unique, something is wrong:\n  name = %s\n  table = %s\n" % (name, flat_tables))

                            # Store table with unique name
                            flat_tables[name] = psi_table

            # Increase unique numpoints*element counter
            counter += 1

    return (element_map, flat_tables)


def unique_psi_tables(tables, eliminate_zeros):
    """Returns a name map and a dictionary of unique tables. The function
    checks if values in the tables are equal, if this is the case it
    creates a name mapping. It also create additional information
    (depending on which parameters are set) such as if the table
    contains all ones, or only zeros, and a list on non-zero columns.
    unique_tables - {name:values,}.  name_map -
    {original_name:[new_name, non-zero-columns (list), is zero (bool),
    is ones (bool)],}.

    """

    # Get unique tables (from old table utility).
    name_map, inverse_name_map = unique_tables(tables)

    debug("\ntables: " + str(tables))
    debug("\nname_map: " + str(name_map))
    debug("\ninv_name_map: " + str(inverse_name_map))

    # Set values to zero if they are lower than threshold.
    format_epsilon = format["epsilon"]
    for name in tables:
        # Get values.
        vals = tables[name]
        for r in range(numpy.shape(vals)[0]):
            for c in range(numpy.shape(vals)[1]):
                if abs(vals[r][c]) < format_epsilon:
                    vals[r][c] = 0
        tables[name] = vals

    # Extract the column numbers that are non-zero.  If optimisation
    # option is set counter for non-zero column arrays.
    i = 0
    non_zero_columns = {}
    if eliminate_zeros:
        for name in sorted(tables.keys()):

            # Get values.
            vals = tables[name]

            # Skip if values are missing
            if len(vals) == 0:
                continue

            # Use the first row as reference.
            non_zeros = list(vals[0].nonzero()[0])

            # If all columns in the first row are non zero, there's no
            # point in continuing.
            if len(non_zeros) == numpy.shape(vals)[1]:
                continue

            # If we only have one row (IP) we just need the nonzero
            # columns.
            if numpy.shape(vals)[0] == 1:
                if list(non_zeros):
                    non_zeros.sort()
                    non_zero_columns[name] = (i, non_zeros)

                    # Compress values.
                    tables[name] = vals[:, non_zeros]
                    i += 1

            # Check if the remaining rows are nonzero in the same
            # positions, else expand.
            else:
                for j in range(1, numpy.shape(vals)[0]):
                    # All rows must have the same non-zero columns for
                    # the optimization to work (at this stage).
                    new_non_zeros = list(vals[j].nonzero()[0])
                    if non_zeros != new_non_zeros:
                        non_zeros = non_zeros + [c for c in new_non_zeros if c not in non_zeros]
                        # If this results in all columns being
                        # non-zero, continue.
                        if len(non_zeros) == numpy.shape(vals)[1]:
                            continue

                # Only add nonzeros if it results in a reduction of
                # columns.
                if len(non_zeros) != numpy.shape(vals)[1]:
                    if list(non_zeros):
                        non_zeros.sort()
                        non_zero_columns[name] = (i, non_zeros)

                        # Compress values.
                        tables[name] = vals[:, non_zeros]
                        i += 1

    # Check if we have some zeros in the tables.
    names_zeros = contains_zeros(tables)

    # Get names of tables with all ones.
    names_ones = get_ones(tables)

    # Add non-zero column, zero and ones info to inverse_name_map (so
    # we only need to pass around one name_map to code generating
    # functions).
    for name in inverse_name_map:
        if inverse_name_map[name] in non_zero_columns:
            nzc = non_zero_columns[inverse_name_map[name]]
            zero = inverse_name_map[name] in names_zeros
            ones = inverse_name_map[name] in names_ones
            inverse_name_map[name] = [inverse_name_map[name], nzc, zero, ones]
        else:
            zero = inverse_name_map[name] in names_zeros
            ones = inverse_name_map[name] in names_ones
            inverse_name_map[name] = [inverse_name_map[name], (), zero, ones]

    # If we found non zero columns we might be able to reduce number
    # of tables further.
    if non_zero_columns:
        # Try reducing the tables. This is possible if some tables
        # have become identical as a consequence of compressing the
        # tables.  This happens with e.g., gradients of linear basis
        # FE0 = {-1,0,1}, nzc0 = [0,2]
        # FE1 = {-1,1,0}, nzc1 = [0,1]  -> FE0 = {-1,1}, nzc0 = [0,2], nzc1 = [0,1].

        # Call old utility function again.
        nm, inv_nm = unique_tables(tables)

        # Update name maps.
        for name in inverse_name_map:
            if inverse_name_map[name][0] in inv_nm:
                inverse_name_map[name][0] = inv_nm[inverse_name_map[name][0]]
        for name in nm:
            maps = nm[name]
            for m in maps:
                if name not in name_map:
                    name_map[name] = []
                if m in name_map:
                    name_map[name] += name_map[m] + [m]
                    del name_map[m]
                else:
                    name_map[name].append(m)

        # Get new names of tables with all ones (for vector
        # constants).
        names = get_ones(tables)

        # Because these tables now contain ones as a consequence of
        # compression we still need to consider the non-zero columns
        # when looking up values in coefficient arrays. The psi
        # entries can however we neglected and we don't need to
        # tabulate the values (if option is set).
        for name in names:
            if name in name_map:
                maps = name_map[name]
                for m in maps:
                    inverse_name_map[m][3] = True
            if name in inverse_name_map:
                    inverse_name_map[name][3] = True

    # Write protect info and return values
    for name in inverse_name_map:
        inverse_name_map[name] = tuple(inverse_name_map[name])

    # Note: inverse_name_map here is called name_map in
    # create_psi_tables and the quadraturetransformerbase class
    return (inverse_name_map, tables)


def unique_tables(tables):
    """Removes tables with redundant values and returns a name_map and a
    inverse_name_map. E.g.,

    tables = {a:[0,1,2], b:[0,2,3], c:[0,1,2], d:[0,1,2]}
    results in:
    tables = {a:[0,1,2], b:[0,2,3]}
    name_map = {a:[c,d]}
    inverse_name_map = {a:a, b:b, c:a, d:a}.

    """

    format_epsilon = format["epsilon"]

    name_map = {}
    inverse_name_map = {}
    names = sorted(tables.keys())
    mapped = []

    # Loop all tables to see if some are redundant.
    for i in range(len(names)):
        name0 = names[i]
        if name0 in mapped:
            continue
        val0 = numpy.array(tables[name0])

        for j in range(i + 1, len(names)):
            name1 = names[j]
            if name1 in mapped:
                continue
            val1 = numpy.array(tables[name1])

            # Check if dimensions match.
            if numpy.shape(val0) == numpy.shape(val1):
                # Check if values are the same.
                if len(val0) > 0 and abs(val0 - val1).max() < format_epsilon:
                    mapped.append(name1)
                    del tables[name1]
                    if name0 in name_map:
                        name_map[name0].append(name1)
                    else:
                        name_map[name0] = [name1]
                    # Create inverse name map.
                    inverse_name_map[name1] = name0

    # Add self.
    for name in tables:
        if name not in inverse_name_map:
            inverse_name_map[name] = name

    return (name_map, inverse_name_map)


def get_ones(tables):
    "Return names of tables for which all values are 1.0."
    f_epsilon = format["epsilon"]
    names = []
    for name in tables:
        vals = tables[name]
        if len(vals) > 0 and abs(vals - numpy.ones(numpy.shape(vals))).max() < f_epsilon:
            names.append(name)
    return names


def contains_zeros(tables):
    "Checks if any tables contains only zeros."
    f_epsilon = format["epsilon"]
    names = []
    for name in tables:
        vals = tables[name]
        if len(vals) > 0 and abs(vals).max() < f_epsilon:
            names.append(name)
    return names


def create_permutations(expr):

    # This is probably not used.
    if len(expr) == 0:
        return expr

    # Format keys and values to lists and tuples.
    if len(expr) == 1:
        new = {}
        for key, val in expr[0].items():
            if key == ():
                pass
            elif not isinstance(key[0], tuple):
                key = (key,)
            if not isinstance(val, list):
                val = [val]
            new[key] = val

        return new

    # Create permutations of two lists.
    # TODO: there could be a cleverer way of changing types of keys and vals.
    if len(expr) == 2:
        new = {}
        for key0, val0 in expr[0].items():
            if isinstance(key0[0], tuple):
                key0 = list(key0)
            if not isinstance(key0, list):
                key0 = [key0]
            if not isinstance(val0, list):
                val0 = [val0]
            for key1, val1 in expr[1].items():
                if key1 == ():
                    key1 = []
                elif isinstance(key1[0], tuple):
                    key1 = list(key1)
                if not isinstance(key1, list):
                    key1 = [key1]
                if not isinstance(val1, list):
                    val1 = [val1]
                if tuple(key0 + key1) in new:
                    error("This is not supposed to happen.")
                new[tuple(key0 + key1)] = val0 + val1

        return new

    # Create permutations by calling this function recursively.  This
    # is only used for rank > 2 tensors I think.
    # if len(expr) > 2:
    #     new = permutations(expr[0:2])
    #     return permutations(new + expr[2:])



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import pytest
import time

# FFC modules
from ffc.quadrature.reduce_operations import operation_count, expand_operations, reduce_operations
from ffc.quadrature.symbolics import *
from ffc.quadrature.sumobj import _group_fractions
from ffc.cpp import format, set_float_formatting
from ffc.parameters import FFC_PARAMETERS
set_float_formatting(FFC_PARAMETERS['precision'])
from ffc.log import error, push_level, pop_level, CRITICAL


def testFractionOperators():
    "Test binary operators"

    f_0 = format["float"](0)
    f_1 = format["float"](1)
    f_2 = format["float"](2)
    f_5 = format["float"](5)

    f2 = FloatValue(2.0)
    fm3 = FloatValue(-3.0)

    x = Symbol("x", GEO)
    y = Symbol("y", GEO)

    p0 = Product([f2, x])
    p1 = Product([x, y])

    S0 = Sum([x, y])

    F0 = Fraction(f2, y)
    F1 = Fraction(x, y)
    F2 = Fraction(x, S0)
    F3 = Fraction(x, y)
    F4 = Fraction(p0, y)
    F5 = Fraction(Product([fm3, x]), y)

    # Test Fraction '+'
    assert str(F0 + f2) == '(%s + %s/y)' % (f_2, f_2)
    assert str(F1 + x) == '(x + x/y)'
    assert str(F1 + p0) == '(%s*x + x/y)' % f_2
    assert str(F1 + S0) == '(x + y + x/y)'
    assert str(F1 + F3) == '%s*x/y' % f_2
    assert str(F0 + F1) == '(%s + x)/y' % f_2
    assert str(F2 + F4) == '(%s*x/y + x/(x + y))' % f_2

    # Test Fraction '-'
    assert str(F0 - f2) == '(%s/y-%s)' % (f_2, f_2)
    assert str(F1 - x) == '(x/y - x)'
    assert str(F1 - p0) == '(x/y-%s*x)' % f_2
    assert str(F1 - S0) == '(x/y - (x + y))'
    assert str(F1 - F3) == '%s' % f_0
    assert str(F4 - F1) == 'x/y'
    assert str(F4 - F5) == '%s*x/y' % f_5
    assert str(F0 - F1) == '(%s - x)/y' % f_2
    assert str(F2 - F4) == '(x/(x + y) - %s*x/y)' % f_2

    # Test Fraction '*'
    assert str(F1 * f2) == '%s*x/y' % f_2
    assert str(F1 * x) == 'x*x/y'
    assert str(F1 * p1) == 'x*x'
    assert str(F1 * S0) == '(x + x*x/y)'
    assert repr(F1 * S0) == repr(Sum([x, Fraction(Product([x, x]), y)]))
    assert str(F1 * F0) == '%s*x/(y*y)' % f_2

    # Test Fraction '/'
    assert str(F0 / f2) == '%s/y' % f_1
    assert str(F1 / x) == '%s/y' % f_1
    assert str(F4 / p1) == '%s/(y*y)' % f_2
    assert str(F4 / x) == '%s/y' % f_2
    assert str(F2 / y) == 'x/(x*y + y*y)'
    assert str(F0 / S0) == '%s/(x*y + y*y)' % f_2

    with pytest.raises(Exception):
        truediv(F0 / F0)



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2016 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

# Python modules
from itertools import chain

# FFC modules
from ffc.log import begin, end, info, error
from ffc.utils import all_equal
from ffc.cpp import make_classname
from ffc.backends.dolfin.wrappers import generate_dolfin_code
from ffc.backends.dolfin.capsules import UFCElementNames, UFCFormNames

__all__ = ["generate_wrapper_code"]

# FIXME: More clean-ups needed here.


def generate_wrapper_code(analysis, prefix, object_names, parameters):
    "Generate code for additional wrappers."

    # Skip if wrappers not requested
    if not parameters["format"] == "dolfin":
        return None

    # Return dolfin wrapper
    return _generate_dolfin_wrapper(analysis, prefix, object_names, parameters)


def _generate_dolfin_wrapper(analysis, prefix, object_names, parameters):

    begin("Compiler stage 4.1: Generating additional wrapper code")

    # Encapsulate data
    (capsules, common_space) = _encapsulate(prefix, object_names, analysis, parameters)

    # Generate code
    info("Generating wrapper code for DOLFIN")
    code = generate_dolfin_code(prefix, "",
                                capsules, common_space,
                                error_control=parameters["error_control"])
    code += "\n\n"
    end()

    return code


def _encapsulate(prefix, object_names, analysis, parameters):

    # Extract data from analysis
    form_datas, elements, element_map, domains = analysis

    # FIXME: Encapsulate domains?

    num_form_datas = len(form_datas)
    common_space = False

    # Special case: single element
    if num_form_datas == 0:
        capsules = _encapsule_element(prefix, elements)

    # Special case: with error control
    elif parameters["error_control"] and num_form_datas == 11:
        capsules = [_encapsule_form(prefix, object_names, form_data, i, element_map)
                    for (i, form_data) in enumerate(form_datas[:num_form_datas - 1])]
        capsules += [_encapsule_form(prefix, object_names, form_datas[-1], num_form_datas - 1,
                                     element_map, "GoalFunctional")]
    # Otherwise: generate standard capsules for each form
    else:
        capsules = [_encapsule_form(prefix, object_names, form_data, i, element_map) for
                    (i, form_data) in enumerate(form_datas)]
        # Check if all argument elements are equal
        elements = []
        for form_data in form_datas:
            elements += form_data.argument_elements
        common_space = all_equal(elements)

    return (capsules, common_space)


def _encapsule_form(prefix, object_names, form_data, i, element_map, superclassname=None):
    element_numbers = [element_map[e] for e in form_data.argument_elements + form_data.coefficient_elements]

    if superclassname is None:
        superclassname = "Form"

    form_names = UFCFormNames(
        object_names.get(id(form_data.original_form), "%d" % i),
        [object_names.get(id(obj), "w%d" % j) for j, obj in enumerate(form_data.reduced_coefficients)],
        make_classname(prefix, "form", i),
        [make_classname(prefix, "finite_element", j) for j in element_numbers],
        [make_classname(prefix, "dofmap", j) for j in element_numbers],
        superclassname)

    return form_names


def _encapsule_element(prefix, elements):
    element_number = len(elements) - 1  # eh? this doesn't make any sense
    args = ("0",
            [make_classname(prefix, "finite_element", element_number)],
            [make_classname(prefix, "dofmap", element_number)])
    return UFCElementNames(*args)



# -*- coding: utf-8 -*-
# Copyright (C) 2013-2016 Martin Sandve Alnæs
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.

import numpy

from ufl.algorithms import replace
from ufl.utils.sorting import sorted_by_count
from ufl import custom_integral_types

from ffc.log import info
from ffc.representationutils import initialize_integral_ir
from ffc.fiatinterface import create_element
from ffc.uflacs.tools import collect_quadrature_rules, compute_quadrature_rules, accumulate_integrals
from ffc.uflacs.build_uflacs_ir import build_uflacs_ir


def compute_integral_ir(itg_data,
                        form_data,
                        form_id,
                        element_numbers,
                        classnames,
                        parameters):
    "Compute intermediate represention of integral."

    info("Computing uflacs representation")

    # Initialise representation
    ir = initialize_integral_ir("uflacs", itg_data, form_data, form_id)

    # Store element classnames
    ir["classnames"] = classnames

    # Get element space dimensions
    unique_elements = element_numbers.keys()
    ir["element_dimensions"] = { ufl_element: create_element(ufl_element).space_dimension()
                                 for ufl_element in unique_elements }

    # Create dimensions of primary indices, needed to reset the argument 'A'
    # given to tabulate_tensor() by the assembler.
    argument_dimensions = [ir["element_dimensions"][ufl_element]
                           for ufl_element in form_data.argument_elements]

    # Compute shape of element tensor
    if ir["integral_type"] == "interior_facet":
        ir["tensor_shape"] = [2 * dim for dim in argument_dimensions]
    else:
        ir["tensor_shape"] = argument_dimensions

    integral_type = itg_data.integral_type
    cell = itg_data.domain.ufl_cell()

    if integral_type in custom_integral_types:
        # Set quadrature degree to twice the highest element degree, to get
        # enough points to identify basis functions via table computations
        max_element_degree = max([1] + [ufl_element.degree() for ufl_element in unique_elements])
        rules = [("default", 2*max_element_degree)]
        quadrature_integral_type = "cell"
    else:
        # Collect which quadrature rules occur in integrals
        default_scheme = itg_data.metadata["quadrature_degree"]
        default_degree = itg_data.metadata["quadrature_rule"]
        rules = collect_quadrature_rules(
            itg_data.integrals, default_scheme, default_degree)
        quadrature_integral_type = integral_type

    # Compute actual points and weights
    quadrature_rules, quadrature_rule_sizes = compute_quadrature_rules(
        rules, quadrature_integral_type, cell)

    # Store quadrature rules in format { num_points: (points, weights) }
    ir["quadrature_rules"] = quadrature_rules

    # Store the fake num_points for analysis in custom integrals
    if integral_type in custom_integral_types:
        ir["fake_num_points"], = quadrature_rules.keys()

    # Group and accumulate integrals on the format { num_points: integral data }
    sorted_integrals = accumulate_integrals(itg_data, quadrature_rule_sizes)

    # Build coefficient numbering for UFC interface here, to avoid
    # renumbering in UFL and application of replace mapping
    if True:
        # Using the mapped coefficients, numbered by UFL
        coefficient_numbering = {}
        sorted_coefficients = sorted_by_count(form_data.function_replace_map.keys())
        for i, f in enumerate(sorted_coefficients):
            g = form_data.function_replace_map[f]
            coefficient_numbering[g] = i
            assert i == g.count()

        # Replace coefficients so they all have proper element and domain for what's to come
        # TODO: We can avoid the replace call when proper Expression support is in place
        #       and element/domain assignment is removed from compute_form_data.
        integrands = {
            num_points: replace(sorted_integrals[num_points].integrand(), form_data.function_replace_map)
            for num_points in sorted(sorted_integrals)
            }
    else:
        pass
        #coefficient_numbering = {}
        #coefficient_element = {}
        #coefficient_domain = {}
        #sorted_coefficients = sorted_by_count(form_data.function_replace_map.keys())
        #for i, f in enumerate(sorted_coefficients):
        #    g = form_data.function_replace_map[f]
        #    coefficient_numbering[f] = i
        #    coefficient_element[f] = g.ufl_element()
        #    coefficient_domain[f] = g.ufl_domain()
        #integrands = {
        #    num_points: sorted_integrals[num_points].integrand()
        #    for num_points in sorted(sorted_integrals)
        #    }
        # then pass coefficient_element and coefficient_domain to the uflacs ir as well


    # Build the more uflacs-specific intermediate representation
    uflacs_ir = build_uflacs_ir(itg_data.domain.ufl_cell(),
                                itg_data.integral_type,
                                ir["entitytype"],
                                integrands,
                                ir["tensor_shape"],
                                coefficient_numbering,
                                quadrature_rules,
                                parameters)
    ir.update(uflacs_ir)

    return ir



# -*- coding: utf-8 -*-
# Copyright (C) 2013 Kristian B. Oelgaard
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Marie E. Rognes, 2013
# Modified by Martin Sandve Alnæs, 2013-2014

from ufl.utils.sorting import sorted_by_key

# FFC modules
from ffc.log import info, error
from ffc.cpp import format
from ffc.quadrature.symbolics import optimise_code, BASIS, IP, GEO
from ffc.quadrature.symbolics import create_product, create_sum, create_symbol, create_fraction


def optimize_integral_ir(ir, parameters):
    "Compute optimized intermediate representation of integral."

    # FIXME: input argument "parameters" has been added to optimize_integral_ir
    # FIXME: which shadows a local parameter

    # Get integral type and optimization parameters
    integral_type = ir["integral_type"]
    parameters = ir["optimise_parameters"]

    # Check whether we should optimize
    if parameters["optimisation"]:

        # Get parameters
        integrals = ir["trans_integrals"]
        integral_type = ir["integral_type"]
        num_facets = ir["num_facets"]
        num_vertices = ir["num_vertices"]
        geo_consts = ir["geo_consts"]
        psi_tables_map = ir["psi_tables_map"]

        # Optimize based on integral type
        if integral_type == "cell":
            info("Optimising expressions for cell integral")
            if parameters["optimisation"] in ("precompute_ip_const", "precompute_basis_const"):
                _precompute_expressions(integrals, geo_consts, parameters["optimisation"])
            else:
                _simplify_expression(integrals, geo_consts, psi_tables_map)
        elif integral_type == "exterior_facet":
            for i in range(num_facets):
                info("Optimising expressions for facet integral %d" % i)
                if parameters["optimisation"] in ("precompute_ip_const", "precompute_basis_const"):
                    _precompute_expressions(integrals[i], geo_consts, parameters["optimisation"])
                else:
                    _simplify_expression(integrals[i], geo_consts, psi_tables_map)
        elif integral_type == "interior_facet":
            for i in range(num_facets):
                for j in range(num_facets):
                    info("Optimising expressions for facet integral (%d, %d)" % (i, j))
                    if parameters["optimisation"] in ("precompute_ip_const", "precompute_basis_const"):
                        _precompute_expressions(integrals[i][j], geo_consts, parameters["optimisation"])
                    else:
                        _simplify_expression(integrals[i][j], geo_consts, psi_tables_map)
        elif integral_type == "vertex":
            for i in range(num_vertices):
                info("Optimising expressions for poin integral %d" % i)
                if parameters["optimisation"] in ("precompute_ip_const", "precompute_basis_const"):
                    _precompute_expressions(integrals[i], geo_consts, parameters["optimisation"])
                else:
                    _simplify_expression(integrals[i], geo_consts, psi_tables_map)
        else:
            error("Unhandled domain type: " + str(integral_type))

    return ir


def _simplify_expression(integral, geo_consts, psi_tables_map):
    for points, terms, functions, ip_consts, coordinate, conditionals in integral:
        # NOTE: sorted is needed to pass the regression tests on the buildbots
        # but it might be inefficient for speed.
        # A solution could be to only compare the output of evaluating the
        # integral, not the header files.
        for loop, (data, entry_vals) in sorted_by_key(terms):
            t_set, u_weights, u_psi_tables, u_nzcs, basis_consts = data
            new_entry_vals = []
            psi_tables = set()
            # NOTE: sorted is needed to pass the regression tests on the buildbots
            # but it might be inefficient for speed.
            # A solution could be to only compare the output of evaluating the
            # integral, not the header files.
            for entry, val, ops in sorted(entry_vals):
                value = optimise_code(val, ip_consts, geo_consts, t_set)
                # Check if value is zero
                if value.val:
                    new_entry_vals.append((entry, value, value.ops()))
                    psi_tables.update(set([psi_tables_map[b] for b in value.get_unique_vars(BASIS)]))

            terms[loop][0][2] = psi_tables
            terms[loop][1] = new_entry_vals


def _precompute_expressions(integral, geo_consts, optimisation):
    for points, terms, functions, ip_consts, coordinate, conditionals in integral:
        for loop, (data, entry_vals) in sorted_by_key(terms):
            t_set, u_weights, u_psi_tables, u_nzcs, basis_consts = data
            new_entry_vals = []
            for entry, val, ops in entry_vals:
                value = _extract_variables(val, basis_consts, ip_consts, geo_consts, t_set, optimisation)
                # Check if value is zero
                if value.val:
                    new_entry_vals.append((entry, value, value.ops()))
            terms[loop][1] = new_entry_vals


def _extract_variables(val, basis_consts, ip_consts, geo_consts, t_set, optimisation):
    f_G = format["geometry constant"]
    f_I = format["ip constant"]
    f_B = format["basis constant"]

    if val._prec == 0:
        return val
    elif val._prec == 1:
        if val.base_expr is None:
            return val
        new_base = _extract_variables(val.base_expr, basis_consts, ip_consts, geo_consts, t_set, optimisation)
        new_sym = create_symbol(val.v, val.t, new_base, val.base_op)
        if new_sym.t == BASIS:
            return _reduce_expression(new_sym, [], basis_consts, f_B, True)
        elif new_sym.t == IP:
            return _reduce_expression(new_sym, [], ip_consts, f_I, True)
        elif new_sym.t == GEO:
            return _reduce_expression(new_sym, [], geo_consts, f_G, True)
    # First handle child classes of product and sum.
    elif val._prec in (2, 3):
        new_vars = []
        for v in val.vrs:
            new_vars.append(_extract_variables(v, basis_consts, ip_consts, geo_consts, t_set, optimisation))
        if val._prec == 2:
            new_val = create_product(new_vars)
        if val._prec == 3:
            new_val = create_sum(new_vars)
    elif val._prec == 4:
        num = _extract_variables(val.num, basis_consts, ip_consts, geo_consts, t_set, optimisation)
        denom = _extract_variables(val.denom, basis_consts, ip_consts, geo_consts, t_set, optimisation)
        return create_fraction(num, denom)
    else:
        error("Unknown symbolic type: %s" % repr(val))

    # Sort variables of product and sum.
    b_c, i_c, g_c = [], [], []
    for v in new_val.vrs:
        if v.t == BASIS:
            if optimisation == "precompute_basis_const":
                b_c.append(v)
        elif v.t == IP:
            i_c.append(v)
        else:
            g_c.append(v)
    vrs = new_val.vrs[:]
    for v in g_c + i_c + b_c:
        vrs.remove(v)
    i_c.extend(_reduce_expression(new_val, g_c, geo_consts, f_G))
    vrs.extend(_reduce_expression(new_val, i_c, ip_consts, f_I))
    vrs.extend(_reduce_expression(new_val, b_c, basis_consts, f_B))

#    print "b_c: "
#    for b in b_c:
#        print b
#    print "basis"
#    for k,v in basis_consts.items():
#        print "k: ", k
#        print "v: ", v
#    print "geo"
#    for k,v in geo_consts.items():
#        print "k: ", k
#        print "v: ", v
#    print "ret val: ", val

    if len(vrs) > 1:
        if new_val._prec == 2:
            new_object = create_product(vrs)
        elif new_val._prec == 3:
            new_object = create_sum(vrs)
        else:
            error("Must have product or sum here: %s" % repr(new_val))
        if new_object.t == BASIS:
            if optimisation == "precompute_ip_const":
                return new_object
            elif optimisation == "precompute_basis_const":
                return _reduce_expression(new_object, [], basis_consts, f_B, True)
        elif new_object.t == IP:
            return _reduce_expression(new_object, [], ip_consts, f_I, True)
        elif new_object.t == GEO:
            return _reduce_expression(new_object, [], geo_consts, f_G, True)
    return vrs[0]

#    if new_val._prec == 2:
#        if len(vrs) > 1:
#            new_prod = create_product(vrs)
#            if new_prod.t == BASIS:
#                if optimisation == "precompute_ip_const":
#                    return new_prod
#                elif optimisation == "precompute_basis_const":
#                    return _reduce_expression(new_prod, [], basis_consts, f_B, True)
#            elif new_prod.t == IP:
#                return _reduce_expression(new_prod, [], ip_consts, f_I, True)
#            elif new_prod.t == GEO:
#                return _reduce_expression(new_prod, [], geo_consts, f_G, True)
#        return vrs[0]
#    elif new_val._prec == 3:
#        if len(vrs) > 1:
#            new_sum = create_sum(vrs)
#            if new_sum.t == BASIS:
#                return new_sum
# return _reduce_expression(new_sum, [], basis_consts, f_B, True)
#            elif new_sum.t == IP:
#                return _reduce_expression(new_sum, [], ip_consts, f_I, True)
#            elif new_sum.t == GEO:
#                return _reduce_expression(new_sum, [], geo_consts, f_G, True)
#        return vrs[0]
#    else:
#        error("Must have product or sum here: %s" % repr(new_val))


def _reduce_expression(expr, symbols, const_dict, f_name, use_expr_type=False):
    if use_expr_type:
        if expr not in const_dict:
            const_dict[expr] = len(const_dict)
        return create_symbol(f_name(const_dict[expr]), expr.t)
    # Only something to be done if we have more than one symbol.
    if len(symbols) > 1:
        sym_type = symbols[0].t
        # Create new symbol.
        if expr._prec == 2:
            new_sym = create_product(symbols)
        elif expr._prec == 3:
            new_sym = create_sum(symbols)
        if new_sym not in const_dict:
            const_dict[new_sym] = len(const_dict)
        s = create_symbol(f_name(const_dict[new_sym]), sym_type)
        return [s]
    return symbols



# -*- coding: utf-8 -*-

# Copyright (C) 2009-2016 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Martin Sandve Alnæs, 2013

"""
Compiler stage 5: optimization
------------------------------

This module implements the optimization of an intermediate code
representation.
"""

# FFC modules
from ffc.log import info, begin, end
from ffc.representation import pick_representation


def optimize_ir(ir, parameters):
    "Optimize intermediate form representation."

    begin("Compiler stage 3: Optimizing intermediate representation")

    # Extract representations
    ir_elements, ir_dofmaps, ir_coordinate_mappings, ir_integrals, ir_forms = ir

    # Check if optimization is requested
    if not any(ir["integrals_metadata"]["optimize"] for ir in ir_integrals):
        info(r"Skipping optimizations, add -O or attach {'optimize': True} "
             "metadata to integrals")

    # Call on every bunch of integrals wich are compiled together
    oir_integrals = [_optimize_integral_ir(ir, parameters)
                     if ir["integrals_metadata"]["optimize"] else ir
                     for ir in ir_integrals]

    end()

    return ir_elements, ir_dofmaps, ir_coordinate_mappings, oir_integrals, ir_forms


def _optimize_integral_ir(ir, parameters):
    "Compute optimized intermediate represention of integral."
    r = pick_representation(ir["representation"])
    return r.optimize_integral_ir(ir, parameters)



# -*- coding: utf-8 -*-
# Code generation format strings for UFC (Unified Form-assembly Code) v. 2017.1.0
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2006-2017

coordinate_mapping_header = """
class %(classname)s: public ufc::coordinate_mapping
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s);

  ~%(classname)s() override;

  const char * signature() const final override;

  coordinate_mapping * create() const final override;

  std::size_t geometric_dimension() const final override;

  std::size_t topological_dimension() const final override;

  ufc::shape cell_shape() const final override;

  ufc::finite_element * create_coordinate_finite_element() const final override;

  ufc::dofmap * create_coordinate_dofmap() const final override;

  void compute_physical_coordinates(
      double * x, std::size_t num_points,
      const double * X,
      const double * coordinate_dofs) const final override;

  void compute_reference_coordinates(
      double * X, std::size_t num_points,
      const double * x,
      const double * coordinate_dofs, double cell_orientation) const final override;

  void compute_jacobians(
      double * J, std::size_t num_points,
      const double * X,
      const double * coordinate_dofs) const final override;

  void compute_jacobian_determinants(
      double * detJ, std::size_t num_points,
      const double * J,
      double cell_orientation) const final override;

  void compute_jacobian_inverses(
      double * K, std::size_t num_points,
      const double * J, const double * detJ) const final override;

  void compute_geometry(
      double * x, double * J, double * detJ, double * K, std::size_t num_points,
      const double * X,
      const double * coordinate_dofs, double cell_orientation) const final override;
};
"""

coordinate_mapping_implementation = """
%(classname)s::%(classname)s(%(constructor_arguments)s) : ufc::coordinate_mapping()%(initializer_list)s
{
%(constructor)s
}

%(classname)s::~%(classname)s() override
{
%(destructor)s
}

const char * %(classname)s::signature() const final override
{
%(signature)s
}

coordinate_mapping * %(classname)s::create() const final override
{
%(create)s
}

std::size_t %(classname)s::geometric_dimension() const final override
{
%(geometric_dimension)s
}

std::size_t %(classname)s::topological_dimension() const final override
{
%(topological_dimension)s
}

ufc::shape %(classname)s::cell_shape() const final override
{
%(cell_shape)s
}

ufc::finite_element * %(classname)s::create_coordinate_finite_element() const final override
{
%(create_coordinate_finite_element)s
}

ufc::dofmap * %(classname)s::create_coordinate_dofmap() const final override
{
%(create_coordinate_dofmap)s
}

void %(classname)s::compute_physical_coordinates(
    double * x, std::size_t num_points,
    const double * X,
    const double * coordinate_dofs) const final override
{
%(compute_physical_coordinates)s
}

void %(classname)s::compute_reference_coordinates(
    double * X, std::size_t num_points,
    const double * x,
    const double * coordinate_dofs, double cell_orientation) const final override
{
%(compute_reference_coordinates)s
}

void %(classname)s::compute_jacobians(
    double * J, std::size_t num_points,
    const double * X,
    const double * coordinate_dofs) const final override
{
%(compute_jacobians)s
}

void %(classname)s::compute_jacobian_determinants(
    double * detJ, std::size_t num_points,
    const double * J,
    double cell_orientation) const final override
{
%(compute_jacobian_determinants)s
}

void %(classname)s::compute_jacobian_inverses(
    double * K, std::size_t num_points,
    const double * J, const double * detJ) const final override
{
%(compute_jacobian_inverses)s
}

void %(classname)s::compute_geometry(
    double * x, double * J, double * detJ, double * K, std::size_t num_points,
    const double * X,
    const double * coordinate_dofs, double cell_orientation) const final override
{
%(compute_geometry)s
}
"""

coordinate_mapping_combined = """
class %(classname)s: public ufc::coordinate_mapping
{%(members)s
public:

  %(classname)s(%(constructor_arguments)s) : ufc::coordinate_mapping()%(initializer_list)s
  {
%(constructor)s
  }

  ~%(classname)s() override
  {
%(destructor)s
  }

  const char * signature() const final override
  {
%(signature)s
  }

  coordinate_mapping * create() const final override
  {
%(create)s
  }

  std::size_t geometric_dimension() const final override
  {
%(geometric_dimension)s
  }

  std::size_t topological_dimension() const final override
  {
%(topological_dimension)s
  }

  ufc::shape cell_shape() const final override
  {
%(cell_shape)s
  }

  ufc::finite_element * create_coordinate_finite_element() const final override
  {
%(create_coordinate_finite_element)s
  }

  ufc::dofmap * create_coordinate_dofmap() const final override
  {
%(create_coordinate_dofmap)s
  }

  void compute_physical_coordinates(
      double * x, std::size_t num_points,
      const double * X,
      const double * coordinate_dofs) const final override
  {
%(compute_physical_coordinates)s
  }

  void compute_reference_coordinates(
      double * X, std::size_t num_points,
      const double * x,
      const double * coordinate_dofs, double cell_orientation) const final override
  {
%(compute_reference_coordinates)s
  }

  void compute_jacobians(
      double * J, std::size_t num_points,
      const double * X,
      const double * coordinate_dofs) const final override
  {
%(compute_jacobians)s
  }

  void compute_jacobian_determinants(
      double * detJ, std::size_t num_points,
      const double * J,
      double cell_orientation) const final override
  {
%(compute_jacobian_determinants)s
  }

  void compute_jacobian_inverses(
      double * K, std::size_t num_points,
      const double * J, const double * detJ) const final override
  {
%(compute_jacobian_inverses)s
  }

  void compute_geometry(
      double * x, double * J, double * detJ, double * K, std::size_t num_points,
      const double * X,
      const double * coordinate_dofs, double cell_orientation) const final override
  {
%(compute_geometry)s
  }

};
"""



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2010-01-25
# Last changed: 2014-04-15

# FFC modules
from ffc.log import debug, error

# FFC tensor representation modules
from ffc.tensor.monomialextraction import extract_monomial_form
from ffc.tensor.monomialtransformation import transform_monomial_form

def estimate_cost(integral, function_replace_map):
    """
    Estimate cost of tensor representation for integral. The cost is
    computed as the sum of the number of coefficients and derivatives,
    if the integrand can be represented as a monomial, and -1 if not.
    """

    # Check that integral type is supported
    supported = ["cell", "exterior_facet", "interior_facet"]
    if not integral.integral_type() in supported:
        return -1

    # Extract monomial integrand
    integrand = integral.integrand()
    try:
        monomial_form = extract_monomial_form([integrand], function_replace_map)
        transform_monomial_form(monomial_form)
    except Exception as exception:
        debug("Monomial extraction failed: " + str(exception))
        return -1

    # Check that we get just one integrand
    if not len(monomial_form) == 1:
        error("Expecting just one integrand.")

    # Compute cost
    cost = 0
    for integrand in monomial_form:
        for monomial in integrand.monomials:
            cost = max(cost, len(monomial.coefficients) + len(monomial.transforms))
    return cost



# -*- coding: utf-8 -*-
"""This module provides functions used by the FFC implementation to
output messages. These may be redirected by the user of FFC.

This module reuses the corresponding log.py module from UFL which
is a wrapper for the standard Python logging module.
"""

# Copyright (C) 2009 Anders Logg
#
# This file is part of FFC.
#
# FFC is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFC. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kristian B. Oelgaard, 2009

# UFL modules
from ufl.log import Logger
from ufl.log import log_functions
from ufl.log import INFO, DEBUG, WARNING, ERROR, CRITICAL
from ufl.utils.sorting import sorted_by_key
from ufl.utils.formatting import dstr, tstr

# Create FFC logger
ffc_logger = Logger("FFC")

# Create FFC global log functions
for foo in log_functions:
    exec("%s = lambda *message : ffc_logger.%s(*message)" % (foo, foo))


# Assertion, copied from UFL
def ffc_assert(condition, *message):
    "Assert that condition is true and otherwise issue an error with given message."
    condition or error(*message)


# Set default log level
set_level(INFO)

#--- Specialized FFC debugging tools ---


def debug_dict(d, title=""):
    "Pretty-print dictionary."
    if not title:
        title = "Dictionary"
    info("")
    begin(title)
    info("")
    for (key, value) in sorted_by_key(d):
        info(key)
        info("-" * len(key))
        info(str(value))
        info("")
    end()


def debug_ir(ir, name=""):
    "Debug intermediate representation."
    title = "Intermediate representation"
    if name:
        title += " (%s)" % str(name)
    debug_dict(ir, title)


def debug_code(code, name=""):
    "Debug generated code."
    title = "Generated code"
    if name:
        title += " (%s)" % str(name)
    debug_dict(code, title)
